# Modified from https://github.com/Lifelong-Robot-Learning/LIBERO/blob/master/libero/libero/envs/venv.py

import cloudpickle
import ctypes
import gymnasium as gym
import numpy as np
import warnings
import time

from abc import ABC, abstractmethod
from collections import OrderedDict
import multiprocessing
from multiprocessing import Array, Pipe, connection
from multiprocessing.context import Process
from typing import Any, Callable, List, Optional, Tuple, Union


# Ref: https://github.com/Lifelong-Robot-Learning/LIBERO/issues/3
# TODO: test it
if multiprocessing.get_start_method(allow_none=True) != "spawn":  
    multiprocessing.set_start_method("spawn", force=True)


gym_old_venv_step_type = Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
gym_new_venv_step_type = Tuple[
    np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray
]
warnings.simplefilter("once", DeprecationWarning)
_NP_TO_CT = {
    np.bool_: ctypes.c_bool,
    np.uint8: ctypes.c_uint8,
    np.uint16: ctypes.c_uint16,
    np.uint32: ctypes.c_uint32,
    np.uint64: ctypes.c_uint64,
    np.int8: ctypes.c_int8,
    np.int16: ctypes.c_int16,
    np.int32: ctypes.c_int32,
    np.int64: ctypes.c_int64,
    np.float32: ctypes.c_float,
    np.float64: ctypes.c_double,
}


def deprecation(msg: str) -> None:
    """Deprecation warning wrapper."""
    warnings.warn(msg, category=DeprecationWarning, stacklevel=2)


class CloudpickleWrapper(object):
    """A cloudpickle wrapper used in SubprocVectorEnv."""

    def __init__(self, data: Any) -> None:
        self.data = data

    def __getstate__(self) -> str:
        return cloudpickle.dumps(self.data)

    def __setstate__(self, data: str) -> None:
        self.data = cloudpickle.loads(data)


GYM_RESERVED_KEYS = [
    "metadata",
    "reward_range",
    "spec",
    "action_space",
    "observation_space",
]


################################################################################
#
# Workers
#
################################################################################


class EnvWorker(ABC):
    """An abstract worker for an environment."""

    def __init__(self, env_fn: Callable[[], gym.Env]) -> None:
        self._env_fn = env_fn
        self.is_closed = False
        self.result: Union[
            gym_old_venv_step_type,
            gym_new_venv_step_type,
            Tuple[np.ndarray, dict],
            np.ndarray,
        ]
        # self.action_space = self.get_env_attr("action_space")  # noqa: B009
        self.is_reset = False

    @abstractmethod
    def get_env_attr(self, key: str) -> Any:
        pass

    @abstractmethod
    def set_env_attr(self, key: str, value: Any) -> None:
        pass

    def send(self, action: Optional[np.ndarray]) -> None:
        """Send action signal to low-level worker.

        When action is None, it indicates sending "reset" signal; otherwise
        it indicates "step" signal. The paired return value from "recv"
        function is determined by such kind of different signal.
        """
        if hasattr(self, "send_action"):
            deprecation(
                "send_action will soon be deprecated. "
                "Please use send and recv for your own EnvWorker."
            )
            if action is None:
                self.is_reset = True
                self.result = self.reset()
            else:
                self.is_reset = False
                self.send_action(action)

    def recv(
        self,
    ) -> Union[
        gym_old_venv_step_type,
        gym_new_venv_step_type,
        Tuple[np.ndarray, dict],
        np.ndarray,
    ]:  # noqa:E125
        """Receive result from low-level worker.

        If the last "send" function sends a NULL action, it only returns a
        single observation; otherwise it returns a tuple of (obs, rew, done,
        info) or (obs, rew, terminated, truncated, info), based on whether
        the environment is using the old step API or the new one.
        """
        if hasattr(self, "get_result"):
            deprecation(
                "get_result will soon be deprecated. "
                "Please use send and recv for your own EnvWorker."
            )
            if not self.is_reset:
                self.result = self.get_result()
        return self.result

    @abstractmethod
    def reset(self, **kwargs: Any) -> Union[np.ndarray, Tuple[np.ndarray, dict]]:
        pass

    def step(
        self, action: np.ndarray
    ) -> Union[gym_old_venv_step_type, gym_new_venv_step_type]:
        """Perform one timestep of the environment's dynamic.

        "send" and "recv" are coupled in sync simulation, so users only call
        "step" function. But they can be called separately in async
        simulation, i.e. someone calls "send" first, and calls "recv" later.
        """
        self.send(action)
        return self.recv()  # type: ignore

    @staticmethod
    def wait(
        workers: List["EnvWorker"], wait_num: int, timeout: Optional[float] = None
    ) -> List["EnvWorker"]:
        """Given a list of workers, return those ready ones."""
        raise NotImplementedError

    def seed(self, seed: Optional[int] = None) -> Optional[List[int]]:
        # return self.action_space.seed(seed)  # issue 299
        pass

    @abstractmethod
    def render(self, **kwargs: Any) -> Any:
        """Render the environment."""
        pass

    @abstractmethod
    def close_env(self) -> None:
        pass

    def close(self) -> None:
        if self.is_closed:
            return None
        self.is_closed = True
        self.close_env()


class ShArray:
    """Wrapper of multiprocessing Array."""

    def __init__(self, dtype: np.generic, shape: Tuple[int]) -> None:
        self.arr = Array(_NP_TO_CT[dtype.type], int(np.prod(shape)))  # type: ignore
        self.dtype = dtype
        self.shape = shape

    def save(self, ndarray: np.ndarray) -> None:
        assert isinstance(ndarray, np.ndarray)
        dst = self.arr.get_obj()
        dst_np = np.frombuffer(dst, dtype=self.dtype).reshape(
            self.shape
        )  # type: ignore
        np.copyto(dst_np, ndarray)

    def get(self) -> np.ndarray:
        obj = self.arr.get_obj()
        return np.frombuffer(obj, dtype=self.dtype).reshape(self.shape)  # type: ignore


def _setup_buf(space: gym.Space) -> Union[dict, tuple, ShArray]:
    if isinstance(space, gym.spaces.Dict):
        assert isinstance(space.spaces, OrderedDict)
        return {k: _setup_buf(v) for k, v in space.spaces.items()}
    elif isinstance(space, gym.spaces.Tuple):
        assert isinstance(space.spaces, tuple)
        return tuple([_setup_buf(t) for t in space.spaces])
    else:
        return ShArray(space.dtype, space.shape)  # type: ignore


def _worker(
    parent: connection.Connection,
    p: connection.Connection,
    env_fn_wrapper: CloudpickleWrapper,
    obs_bufs: Optional[Union[dict, tuple, ShArray]] = None,
) -> None:
    def _encode_obs(
        obs: Union[dict, tuple, np.ndarray], buffer: Union[dict, tuple, ShArray]
    ) -> None:
        if isinstance(obs, np.ndarray) and isinstance(buffer, ShArray):
            buffer.save(obs)
        elif isinstance(obs, tuple) and isinstance(buffer, tuple):
            for o, b in zip(obs, buffer):
                _encode_obs(o, b)
        elif isinstance(obs, dict) and isinstance(buffer, dict):
            for k in obs.keys():
                _encode_obs(obs[k], buffer[k])
        return None

    parent.close()
    env = env_fn_wrapper.data()
    try:
        while True:
            try:
                cmd, data = p.recv()
            except EOFError:  # the pipe has been closed
                p.close()
                break
            if cmd == "step":
                env_return = env.step(data)
                if obs_bufs is not None:
                    _encode_obs(env_return[0], obs_bufs)
                    env_return = (None, *env_return[1:])
                p.send(env_return)
            elif cmd == "reinit_env":
                env.close()
                env = data.data() # data is a CloudpickleWrapper with the new env_fn
                p.send(True)
            elif cmd == "reset":
                retval = env.reset(**data)
                reset_returns_info = (
                    isinstance(retval, (tuple, list))
                    and len(retval) == 2
                    and isinstance(retval[1], dict)
                )
                if reset_returns_info:
                    obs, info = retval
                else:
                    obs = retval
                if obs_bufs is not None:
                    _encode_obs(obs, obs_bufs)
                    obs = None
                if reset_returns_info:
                    p.send((obs, info))
                else:
                    p.send(obs)
            elif cmd == "close":
                p.send(env.close())
                p.close()
                break
            elif cmd == "render":
                p.send(env.render(**data) if hasattr(env, "render") else None)
            elif cmd == "seed":
                if hasattr(env, "seed"):
                    p.send(env.seed(data))
                else:
                    env.reset(seed=data)
                    p.send(None)
            elif cmd == "getattr":
                p.send(getattr(env, data) if hasattr(env, data) else None)
            elif cmd == "setattr":
                setattr(env.unwrapped, data["key"], data["value"])
            elif cmd == "check_success":
                p.send(env.check_success())
            elif cmd == "get_segmentation_of_interest":
                p.send(env.get_segmentation_of_interest(data))
            elif cmd == "get_sim_state":
                p.send(env.get_sim_state())
            elif cmd == "set_init_state":
                obs = env.set_init_state(data)
                p.send(obs)
            else:
                p.close()
                raise NotImplementedError
    except KeyboardInterrupt:
        p.close()


class DummyEnvWorker(EnvWorker):
    """Dummy worker used in sequential vector environments."""

    def __init__(self, env_fn: Callable[[], gym.Env]) -> None:
        self.env = env_fn()
        super().__init__(env_fn)

    def get_env_attr(self, key: str) -> Any:
        return getattr(self.env, key)

    def set_env_attr(self, key: str, value: Any) -> None:
        setattr(self.env.unwrapped, key, value)

    def reset(self, **kwargs: Any) -> Union[np.ndarray, Tuple[np.ndarray, dict]]:
        if "seed" in kwargs:
            super().seed(kwargs["seed"])
        return self.env.reset(**kwargs)

    @staticmethod
    def wait(  # type: ignore
        workers: List["DummyEnvWorker"], wait_num: int, timeout: Optional[float] = None
    ) -> List["DummyEnvWorker"]:
        # Sequential EnvWorker objects are always ready
        return workers

    def send(self, action: Optional[np.ndarray], **kwargs: Any) -> None:
        if action is None:
            self.result = self.env.reset(**kwargs)
        else:
            self.result = self.env.step(action)  # type: ignore

    def seed(self, seed: Optional[int] = None) -> Optional[List[int]]:
        super().seed(seed)
        try:
            return self.env.seed(seed)  # type: ignore
        except (AttributeError, NotImplementedError):
            self.env.reset(seed=seed)
            return [seed]  # type: ignore

    def render(self, **kwargs: Any) -> Any:
        return self.env.render(**kwargs)

    def close_env(self) -> None:
        self.env.close()

    def check_success(self):
        return self.env.check_success()

    def get_segmentation_of_interest(self, segmentation_image):
        return self.env.get_segmentation_of_interest(segmentation_image)

    def get_sim_state(self):
        return self.env.get_sim_state()

    def set_init_state(self, init_state):
        return self.env.set_init_state(init_state)


class SubprocEnvWorker(EnvWorker):
    """Subprocess worker used in SubprocVectorEnv and ShmemVectorEnv."""

    def __init__(
        self, env_fn: Callable[[], gym.Env], share_memory: bool = False
    ) -> None:
        self.parent_remote, self.child_remote = Pipe()
        self.share_memory = share_memory
        self.buffer: Optional[Union[dict, tuple, ShArray]] = None
        if self.share_memory:
            dummy = env_fn()
            obs_space = dummy.observation_space
            dummy.close()
            del dummy
            self.buffer = _setup_buf(obs_space)
        args = (
            self.parent_remote,
            self.child_remote,
            CloudpickleWrapper(env_fn),
            self.buffer,
        )
        self.process = Process(target=_worker, args=args, daemon=True)
        self.process.start()
        self.child_remote.close()
        super().__init__(env_fn)

    def get_env_attr(self, key: str) -> Any:
        self.parent_remote.send(["getattr", key])
        return self.parent_remote.recv()

    def set_env_attr(self, key: str, value: Any) -> None:
        self.parent_remote.send(["setattr", {"key": key, "value": value}])

    def _decode_obs(self) -> Union[dict, tuple, np.ndarray]:
        def decode_obs(
            buffer: Optional[Union[dict, tuple, ShArray]]
        ) -> Union[dict, tuple, np.ndarray]:
            if isinstance(buffer, ShArray):
                return buffer.get()
            elif isinstance(buffer, tuple):
                return tuple([decode_obs(b) for b in buffer])
            elif isinstance(buffer, dict):
                return {k: decode_obs(v) for k, v in buffer.items()}
            else:
                raise NotImplementedError

        return decode_obs(self.buffer)

    @staticmethod
    def wait(  # type: ignore
        workers: List["SubprocEnvWorker"],
        wait_num: int,
        timeout: Optional[float] = None,
    ) -> List["SubprocEnvWorker"]:
        remain_conns = conns = [x.parent_remote for x in workers]
        ready_conns: List[connection.Connection] = []
        remain_time, t1 = timeout, time.time()
        while len(remain_conns) > 0 and len(ready_conns) < wait_num:
            if timeout:
                remain_time = timeout - (time.time() - t1)
                if remain_time <= 0:
                    break
            # connection.wait hangs if the list is empty
            new_ready_conns = connection.wait(remain_conns, timeout=remain_time)
            ready_conns.extend(new_ready_conns)  # type: ignore
            remain_conns = [conn for conn in remain_conns if conn not in ready_conns]
        return [workers[conns.index(con)] for con in ready_conns]
        
    def send_reinit(self, env_fn: Callable[[], gym.Env]) -> None:
        """
        Sends a reinit command without waiting for a response.
        This allows for parallel initialization.
        """
        self.parent_remote.send(("reinit_env", CloudpickleWrapper(env_fn)))

    def recv_reinit(self) -> bool:
        """
        Waits for and receives the confirmation from a reinit command.
        """
        try:
            return self.parent_remote.recv()
        except (BrokenPipeError, EOFError):
            return False

    def send(self, action: Optional[np.ndarray], **kwargs: Any) -> None:
        if action is None:
            if "seed" in kwargs:
                super().seed(kwargs["seed"])
            self.parent_remote.send(["reset", kwargs])
        else:
            self.parent_remote.send(["step", action])

    def recv(
        self,
    ) -> Union[
        gym_old_venv_step_type,
        gym_new_venv_step_type,
        Tuple[np.ndarray, dict],
        np.ndarray,
    ]:  # noqa:E125
        result = self.parent_remote.recv()
        if isinstance(result, tuple):
            if len(result) == 2:
                obs, info = result
                if self.share_memory:
                    obs = self._decode_obs()
                return obs, info
            obs = result[0]
            if self.share_memory:
                obs = self._decode_obs()
            return (obs, *result[1:])  # type: ignore
        else:
            obs = result
            if self.share_memory:
                obs = self._decode_obs()
            return obs

    def reset(self, **kwargs: Any) -> Union[np.ndarray, Tuple[np.ndarray, dict]]:
        if "seed" in kwargs:
            super().seed(kwargs["seed"])
        self.parent_remote.send(["reset", kwargs])

        result = self.parent_remote.recv()
        if isinstance(result, tuple):
            obs, info = result
            if self.share_memory:
                obs = self._decode_obs()
            return obs, info
        else:
            obs = result
            if self.share_memory:
                obs = self._decode_obs()
            return obs

    def seed(self, seed: Optional[int] = None) -> Optional[List[int]]:
        super().seed(seed)
        self.parent_remote.send(["seed", seed])
        ret = self.parent_remote.recv()
        return ret

    def render(self, **kwargs: Any) -> Any:
        self.parent_remote.send(["render", kwargs])
        return self.parent_remote.recv()

    def close_env(self) -> None:
        try:
            self.parent_remote.send(["close", None])
            # mp may be deleted so it may raise AttributeError
            self.parent_remote.recv()
            self.process.join()
        except (BrokenPipeError, EOFError, AttributeError):
            pass
        # ensure the subproc is terminated
        self.process.terminate()

    def check_success(self):
        self.parent_remote.send(["check_success", None])
        return self.parent_remote.recv()

    def get_segmentation_of_interest(self, segmentation_image):
        self.parent_remote.send(["get_segmentation_of_interest", segmentation_image])
        return self.parent_remote.recv()

    def get_sim_state(self):
        self.parent_remote.send(["get_sim_state", None])
        return self.parent_remote.recv()

    def set_init_state(self, init_state):
        self.parent_remote.send(["set_init_state", init_state])
        obs = self.parent_remote.recv()
        if self.share_memory:
            obs = self._decode_obs()
        return obs


################################################################################
#
# VecEnvs
#
################################################################################


class BaseVectorEnv(object):
    """Base class for vectorized environments.

    Usage:
    ::

        env_num = 8
        envs = DummyVectorEnv([lambda: gym.make(task) for _ in range(env_num)])
        assert len(envs) == env_num

    It accepts a list of environment generators. In other words, an environment
    generator ``efn`` of a specific task means that ``efn()`` returns the
    environment of the given task, for example, ``gym.make(task)``.

    All of the VectorEnv must inherit :class:`~tianshou.env.BaseVectorEnv`.
    Here are some other usages:
    ::

        envs.seed(2)  # which is equal to the next line
        envs.seed([2, 3, 4, 5, 6, 7, 8, 9])  # set specific seed for each env
        obs = envs.reset()  # reset all environments
        obs = envs.reset([0, 5, 7])  # reset 3 specific environments
        obs, rew, done, info = envs.step([1] * 8)  # step synchronously
        envs.render()  # render all environments
        envs.close()  # close all environments

    .. warning::

        If you use your own environment, please make sure the ``seed`` method
        is set up properly, e.g.,
        ::

            def seed(self, seed):
                np.random.seed(seed)

        Otherwise, the outputs of these envs may be the same with each other.

    :param env_fns: a list of callable envs, ``env_fns[i]()`` generates the i-th env.
    :param worker_fn: a callable worker, ``worker_fn(env_fns[i])`` generates a
        worker which contains the i-th env.
    :param int wait_num: use in asynchronous simulation if the time cost of
        ``env.step`` varies with time and synchronously waiting for all
        environments to finish a step is time-wasting. In that case, we can
        return when ``wait_num`` environments finish a step and keep on
        simulation in these environments. If ``None``, asynchronous simulation
        is disabled; else, ``1 <= wait_num <= env_num``.
    :param float timeout: use in asynchronous simulation same as above, in each
        vectorized step it only deal with those environments spending time
        within ``timeout`` seconds.
    """

    def __init__(
        self,
        env_fns: List[Callable[[], gym.Env]],
        worker_fn: Callable[[Callable[[], gym.Env]], EnvWorker],
        wait_num: Optional[int] = None,
        timeout: Optional[float] = None,
    ) -> None:
        self._env_fns = env_fns
        # A VectorEnv contains a pool of EnvWorkers, which corresponds to
        # interact with the given envs (one worker <-> one env).
        self.workers = [worker_fn(fn) for fn in env_fns]
        self.worker_class = type(self.workers[0])
        assert issubclass(self.worker_class, EnvWorker)
        assert all([isinstance(w, self.worker_class) for w in self.workers])

        self.env_num = len(env_fns)
        self.wait_num = wait_num or len(env_fns)
        assert (
            1 <= self.wait_num <= len(env_fns)
        ), f"wait_num should be in [1, {len(env_fns)}], but got {wait_num}"
        self.timeout = timeout
        assert (
            self.timeout is None or self.timeout > 0
        ), f"timeout is {timeout}, it should be positive if provided!"
        self.is_async = self.wait_num != len(env_fns) or timeout is not None
        self.waiting_conn: List[EnvWorker] = []
        # environments in self.ready_id is actually ready
        # but environments in self.waiting_id are just waiting when checked,
        # and they may be ready now, but this is not known until we check it
        # in the step() function
        self.waiting_id: List[int] = []
        # all environments are ready in the beginning
        self.ready_id = list(range(self.env_num))
        self.is_closed = False

    def _assert_is_not_closed(self) -> None:
        assert (
            not self.is_closed
        ), f"Methods of {self.__class__.__name__} cannot be called after close."

    def __len__(self) -> int:
        """Return len(self), which is the number of environments."""
        return self.env_num

    def __getattribute__(self, key: str) -> Any:
        """Switch the attribute getter depending on the key.

        Any class who inherits ``gym.Env`` will inherit some attributes, like
        ``action_space``. However, we would like the attribute lookup to go straight
        into the worker (in fact, this vector env's action_space is always None).
        """
        if key in GYM_RESERVED_KEYS:  # reserved keys in gym.Env
            return self.get_env_attr(key)
        else:
            return super().__getattribute__(key)

    def get_env_attr(
        self,
        key: str,
        id: Optional[Union[int, List[int], np.ndarray]] = None,
    ) -> List[Any]:
        """Get an attribute from the underlying environments.

        If id is an int, retrieve the attribute denoted by key from the environment
        underlying the worker at index id. The result is returned as a list with one
        element. Otherwise, retrieve the attribute for all workers at indices id and
        return a list that is ordered correspondingly to id.

        :param str key: The key of the desired attribute.
        :param id: Indice(s) of the desired worker(s). Default to None for all env_id.

        :return list: The list of environment attributes.
        """
        self._assert_is_not_closed()
        id = self._wrap_id(id)
        if self.is_async:
            self._assert_id(id)

        return [self.workers[j].get_env_attr(key) for j in id]

    def set_env_attr(
        self,
        key: str,
        value: Any,
        id: Optional[Union[int, List[int], np.ndarray]] = None,
    ) -> None:
        """Set an attribute in the underlying environments.

        If id is an int, set the attribute denoted by key from the environment
        underlying the worker at index id to value.
        Otherwise, set the attribute for all workers at indices id.

        :param str key: The key of the desired attribute.
        :param Any value: The new value of the attribute.
        :param id: Indice(s) of the desired worker(s). Default to None for all env_id.
        """
        self._assert_is_not_closed()
        id = self._wrap_id(id)
        if self.is_async:
            self._assert_id(id)
        for j in id:
            self.workers[j].set_env_attr(key, value)

    def _wrap_id(
        self,
        id: Optional[Union[int, List[int], np.ndarray]] = None,
    ) -> Union[List[int], np.ndarray]:
        if id is None:
            return list(range(self.env_num))
        return [id] if np.isscalar(id) else id  # type: ignore

    def _assert_id(self, id: Union[List[int], np.ndarray]) -> None:
        for i in id:
            assert (
                i not in self.waiting_id
            ), f"Cannot interact with environment {i} which is stepping now."
            assert (
                i in self.ready_id
            ), f"Can only interact with ready environments {self.ready_id}."

    def reset(
        self,
        id: Optional[Union[int, List[int], np.ndarray]] = None,
        **kwargs: Any,
    ) -> Union[np.ndarray, Tuple[np.ndarray, Union[dict, List[dict]]]]:
        """Reset the state of some envs and return initial observations.

        If id is None, reset the state of all the environments and return
        initial observations, otherwise reset the specific environments with
        the given id, either an int or a list.
        """
        self._assert_is_not_closed()
        id = self._wrap_id(id)
        if self.is_async:
            self._assert_id(id)

        # send(None) == reset() in worker
        for i in id:
            self.workers[i].send(None, **kwargs)
        ret_list = [self.workers[i].recv() for i in id]

        reset_returns_info = (
            isinstance(ret_list[0], (tuple, list))
            and len(ret_list[0]) == 2
            and isinstance(ret_list[0][1], dict)
        )
        if reset_returns_info:
            obs_list = [r[0] for r in ret_list]
        else:
            obs_list = ret_list

        if isinstance(obs_list[0], tuple):
            raise TypeError(
                "Tuple observation space is not supported. ",
                "Please change it to array or dict space",
            )
        try:
            obs = np.stack(obs_list)
        except ValueError:  # different len(obs)
            obs = np.array(obs_list, dtype=object)

        if reset_returns_info:
            infos = [r[1] for r in ret_list]
            return obs, infos  # type: ignore
        else:
            return obs

    def step(
        self,
        action: np.ndarray,
        id: Optional[Union[int, List[int], np.ndarray]] = None,
    ) -> Union[gym_old_venv_step_type, gym_new_venv_step_type]:
        """Run one timestep of some environments' dynamics.

        If id is None, run one timestep of all the environments’ dynamics;
        otherwise run one timestep for some environments with given id,  either
        an int or a list. When the end of episode is reached, you are
        responsible for calling reset(id) to reset this environment’s state.

        Accept a batch of action and return a tuple (batch_obs, batch_rew,
        batch_done, batch_info) in numpy format.

        :param numpy.ndarray action: a batch of action provided by the agent.

        :return: A tuple consisting of either:

            * ``obs`` a numpy.ndarray, the agent's observation of current environments
            * ``rew`` a numpy.ndarray, the amount of rewards returned after \
                previous actions
            * ``done`` a numpy.ndarray, whether these episodes have ended, in \
                which case further step() calls will return undefined results
            * ``info`` a numpy.ndarray, contains auxiliary diagnostic \
                information (helpful for debugging, and sometimes learning)

            or:

            * ``obs`` a numpy.ndarray, the agent's observation of current environments
            * ``rew`` a numpy.ndarray, the amount of rewards returned after \
                previous actions
            * ``terminated`` a numpy.ndarray, whether these episodes have been \
                terminated
            * ``truncated`` a numpy.ndarray, whether these episodes have been truncated
            * ``info`` a numpy.ndarray, contains auxiliary diagnostic \
                information (helpful for debugging, and sometimes learning)

            The case distinction is made based on whether the underlying environment
            uses the old step API (first case) or the new step API (second case).

        For the async simulation:

        Provide the given action to the environments. The action sequence
        should correspond to the ``id`` argument, and the ``id`` argument
        should be a subset of the ``env_id`` in the last returned ``info``
        (initially they are env_ids of all the environments). If action is
        None, fetch unfinished step() calls instead.
        """
        self._assert_is_not_closed()
        id = self._wrap_id(id)
        if not self.is_async:
            assert len(action) == len(id)
            for i, j in enumerate(id):
                self.workers[j].send(action[i])
            result = []
            for j in id:
                env_return = self.workers[j].recv()
                env_return[-1]["env_id"] = j
                result.append(env_return)
        else:
            if action is not None:
                self._assert_id(id)
                assert len(action) == len(id)
                for act, env_id in zip(action, id):
                    self.workers[env_id].send(act)
                    self.waiting_conn.append(self.workers[env_id])
                    self.waiting_id.append(env_id)
                self.ready_id = [x for x in self.ready_id if x not in id]
            ready_conns: List[EnvWorker] = []
            while not ready_conns:
                ready_conns = self.worker_class.wait(
                    self.waiting_conn, self.wait_num, self.timeout
                )
            result = []
            for conn in ready_conns:
                waiting_index = self.waiting_conn.index(conn)
                self.waiting_conn.pop(waiting_index)
                env_id = self.waiting_id.pop(waiting_index)
                # env_return can be (obs, reward, done, info) or
                # (obs, reward, terminated, truncated, info)
                env_return = conn.recv()
                env_return[-1]["env_id"] = env_id  # Add `env_id` to info
                result.append(env_return)
                self.ready_id.append(env_id)
        return_lists = tuple(zip(*result))
        obs_list = return_lists[0]
        try:
            obs_stack = np.stack(obs_list)
        except ValueError:  # different len(obs)
            obs_stack = np.array(obs_list, dtype=object)
        other_stacks = map(np.stack, return_lists[1:])
        return (obs_stack, *other_stacks)  # type: ignore

    def seed(
        self,
        seed: Optional[Union[int, List[int]]] = None,
    ) -> List[Optional[List[int]]]:
        """Set the seed for all environments.

        Accept ``None``, an int (which will extend ``i`` to
        ``[i, i + 1, i + 2, ...]``) or a list.

        :return: The list of seeds used in this env's random number generators.
            The first value in the list should be the "main" seed, or the value
            which a reproducer pass to "seed".
        """
        self._assert_is_not_closed()
        seed_list: Union[List[None], List[int]]
        if seed is None:
            seed_list = [seed] * self.env_num
        elif isinstance(seed, int):
            seed_list = [seed + i for i in range(self.env_num)]
        else:
            seed_list = seed
        return [w.seed(s) for w, s in zip(self.workers, seed_list)]

    def render(self, **kwargs: Any) -> List[Any]:
        """Render all of the environments."""
        self._assert_is_not_closed()
        if self.is_async and len(self.waiting_id) > 0:
            raise RuntimeError(
                f"Environments {self.waiting_id} are still stepping, cannot "
                "render them now."
            )
        return [w.render(**kwargs) for w in self.workers]

    def close(self) -> None:
        """Close all of the environments.

        This function will be called only once (if not, it will be called during
        garbage collected). This way, ``close`` of all workers can be assured.
        """
        self._assert_is_not_closed()
        for w in self.workers:
            w.close()
        self.is_closed = True


class DummyVectorEnv(BaseVectorEnv):
    """Dummy vectorized environment wrapper, implemented in for-loop.

    .. seealso::

        Please refer to :class:`~tianshou.env.BaseVectorEnv` for other APIs' usage.
    """

    def __init__(self, env_fns: List[Callable[[], gym.Env]], **kwargs: Any) -> None:
        super().__init__(env_fns, DummyEnvWorker, **kwargs)

    def check_success(self):
        return [w.check_success() for w in self.workers]

    def get_segmentation_of_interest(self, segmentation_images):
        return [
            w.get_segmentation_of_interest(img)
            for w, img in zip(self.workers, segmentation_images)
        ]

    def get_sim_state(self):
        return [w.get_sim_state() for w in self.workers]

    def set_init_state(
        self,
        init_state: Optional[Union[int, List[int], np.ndarray]] = None,
        id: Optional[Union[int, List[int], np.ndarray]] = None,
        **kwargs: Any,
    ) -> Union[np.ndarray, Tuple[np.ndarray, Union[dict, List[dict]]]]:
        """Reset the state of some envs and return initial observations.
        If id is None, reset the state of all the environments and return
        initial observations, otherwise reset the specific environments with
        the given id, either an int or a list.
        """
        self._assert_is_not_closed()
        id = self._wrap_id(id)
        if self.is_async:
            self._assert_id(id)

        # send(None) == reset() in worker
        obs_list = []
        for j, i in enumerate(id):
            obs = self.workers[i].set_init_state(init_state[j])
            obs_list.append(obs)
        obs = np.stack(obs_list)
        return obs


class SubprocVectorEnv(BaseVectorEnv):
    """Vectorized environment wrapper based on subprocess.

    .. seealso::

        Please refer to :class:`~tianshou.env.BaseVectorEnv` for other APIs' usage.
    """

    def __init__(self, env_fns: List[Callable[[], gym.Env]], **kwargs: Any) -> None:
        def worker_fn(fn: Callable[[], gym.Env]) -> SubprocEnvWorker:
            return SubprocEnvWorker(fn, share_memory=False)

        super().__init__(env_fns, worker_fn, **kwargs)

    def reinit_envs(self, env_fns: List[Callable[[], gym.Env]], id: Optional[Union[int, List[int], np.ndarray]] = None) -> None:
        """
        Re-initializes the environments in parallel for a subset of workers.
        """
        self._assert_is_not_closed()

        target_ids = self._wrap_id(id)
        
        if len(env_fns) != len(target_ids):
            raise ValueError(f"Number of env_fns ({len(env_fns)}) must match the number of target ids ({len(target_ids)}).")

        target_workers = [self.workers[i] for i in target_ids]

        # Send reinit command to all target workers without waiting.
        for i, (worker, env_fn) in enumerate(zip(target_workers, env_fns)):
            if not isinstance(worker, SubprocEnvWorker):
                raise TypeError(
                    f"reinit_envs is only supported for SubprocEnvWorker, but worker {target_ids[i]} is type {type(worker).__name__}."
                )
            worker.send_reinit(env_fn)

        # Wait for all target workers to confirm reinitialization.
        results = [worker.recv_reinit() for worker in target_workers]
        if not all(results):
            failed_indices = [target_ids[i] for i, success in enumerate(results) if not success]
            raise RuntimeError(f"Worker processes {failed_indices} failed to re-initialize environment.")

    def check_success(self):
        return [w.check_success() for w in self.workers]

    def get_segmentation_of_interest(self, segmentation_images):
        return [
            w.get_segmentation_of_interest(img)
            for w, img in zip(self.workers, segmentation_images)
        ]

    def get_sim_state(self):
        return [w.get_sim_state() for w in self.workers]

    def set_init_state(
        self,
        init_state: Optional[Union[int, List[int], np.ndarray]] = None,
        id: Optional[Union[int, List[int], np.ndarray]] = None,
        **kwargs: Any,
    ) -> Union[np.ndarray, Tuple[np.ndarray, Union[dict, List[dict]]]]:
        """Reset the state of some envs and return initial observations.
        If id is None, reset the state of all the environments and return
        initial observations, otherwise reset the specific environments with
        the given id, either an int or a list.
        """
        self._assert_is_not_closed()
        id = self._wrap_id(id)
        if self.is_async:
            self._assert_id(id)

        # send(None) == reset() in worker
        obs_list = []
        for j, i in enumerate(id):
            obs = self.workers[i].set_init_state(init_state[j])
            obs_list.append(obs)
        obs = np.stack(obs_list)
        return obs




# Copyright 2025, Shanghai Innovation Institute. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import uuid
import ray
import torch
import asyncio
import numpy as np
import torch.distributed as dist
from collections import defaultdict
from pprint import pformat
from tqdm import tqdm
from loguru import logger
from typing import Any, Dict, List, Optional, Set, Tuple, Type, Callable
from torch.distributed import ProcessGroup
from tensordict import TensorDict
from tensordict.tensorclass import NonTensorData
import time
from siirl.execution.metric_worker.metric_worker import MetricClient
from siirl.models.loader import TokenizerModule, load_tokenizer
from siirl.params import SiiRLArguments
from siirl.engine.base_worker import Worker
from siirl.execution.dag import TaskGraph
from siirl.execution.dag.node import NodeRole, NodeType, Node
from siirl.execution.scheduler.reward import compute_reward, create_reward_manager
from siirl.execution.scheduler.process_group_manager import ProcessGroupManager
from siirl.execution.scheduler.enums import AdvantageEstimator, WorkflowType
from siirl.data_coordinator import preprocess_dataloader, Samples2Dict, Dict2Samples, SampleInfo
from siirl.data_coordinator.dataloader import DataLoaderNode
from siirl.dag_worker.data_structures import NodeOutput
from siirl.dag_worker.constants import DAGConstants, DAGInitializationError
from siirl.dag_worker import core_algos
from siirl.dag_worker.checkpoint_manager import CheckpointManager
from siirl.dag_worker.core_algos import (
    agg_loss,
    apply_kl_penalty,
    compute_advantage,
    compute_response_mask
    )
from siirl.dag_worker.dag_utils import  (
    log_ray_actor_info,
    get_and_validate_rank,
    get_taskgraph_for_rank,
    log_role_worker_mapping,
    should_create_worker,
    generate_node_worker_key,
    find_first_non_compute_ancestor,
    setup_sharding_manager,
    get_worker_classes,
    get_parallelism_config,
    prepare_generation_batch,
    format_metrics_by_group,
    log_metrics_to_console,
    aggregate_and_write_performance_metrics,
    log_core_performance_metrics,
    timer,
    reduce_and_broadcast_metrics,
    whether_put_data
    )
from siirl.utils.debug import DistProfiler
from siirl.utils.extras.device import get_device_name, get_nccl_backend
from siirl.execution.rollout_flow.multiturn.agent_loop import AgentLoopManager

device_name = get_device_name()

class DAGWorker(Worker):
    """
    Orchestrates a Directed Acyclic Graph (DAG) of tasks for distributed training,
    managing the setup, initialization, and workflow for a specific rank.
    """

    def __init__(
        self,
        config: SiiRLArguments,
        process_group_manager: ProcessGroupManager,
        taskgraph_mapping: Dict[int, TaskGraph],
        data_coordinator: "ray.actor.ActorHandle",
        metric_worker: "ray.actor.ActorHandle",
        device_name="cuda",
    ):
        super().__init__()
        self.config = config
        self.process_group_manager = process_group_manager
        self.taskgraph_mapping = taskgraph_mapping
        self.data_coordinator = data_coordinator
        self.device_name = device_name
        self.enable_perf = os.environ.get("SIIRL_ENABLE_PERF", "0") == "1" or config.dag.enable_perf

        # State attributes
        self.timing_raw = {}
        self.global_steps = 0
        self.total_training_steps = 0
        self.workers: Dict[str, Any] = {}
        self.multi_agent_group: Dict[int, Dict[NodeRole, Any]] = defaultdict(dict)
        self.agent_group_process_group: Dict[int, Dict[NodeRole, Any]] = defaultdict(dict)
        self.process_groups: Dict[str, ProcessGroup] = {}
        self.tokenizer_mapping: Dict[str, TokenizerModule] = {}
        self.logger = None
        self.progress_bar = None
        self._rank: int = -1
        self.taskgraph: Optional[TaskGraph] = None
        self.internal_data_cache: Dict[str, Any] = {}
        self.sample_ref_cache: list = []
        self.agent_critic_worker: Any
        # Finish flag
        self.taskgraph_execute_finished = False

        # async rollout
        self.rollout_mode = "sync"
        self._async_rollout_manager = None
        self.zmq_address = None  # used for async_vllmrollout

        # Add a cache to hold data from an insufficient batch for the next training step.
        # This is the core state-carrying mechanism for dynamic sampling.
        self.sampling_leftover_cache: Optional[Any] = None

        # multi agent
        self._multi_agent = False
        
        # metirc_worker
        self.metric_worker = MetricClient(metric_worker=metric_worker)
        try:
            self._initialize_worker()
        except (ValueError, TypeError, KeyError, AttributeError, NotImplementedError) as e:
            rank = os.environ.get("RANK", "UNKNOWN")
            logger.error(f"Rank {rank}: Failed to create DAGWorker due to a critical setup error: {e}", exc_info=True)
            raise DAGInitializationError(f"Initialization failed on Rank {rank}: {e}") from e

        log_ray_actor_info(self._rank)

# ==========================================================================================
# Module 1: Execution and Training Loop
# ==========================================================================================

    def execute_task_graph(self):
        """Main entry point to start the DAG execution pipeline."""
        logger.info(f"Rank {self._rank}: Starting DAG execution pipeline...")
        logger.success(f"Rank {self._rank}: All components initialized. Starting training loop from step {self.global_steps + 1}.")

        if self.config.trainer.val_before_train:
            self.validator.validate(global_step=self.global_steps)
            self.metric_worker.wait_submit()
            dist.barrier(self._gather_group)
            if self._rank == 0 and self.logger:
                val_metrics = self.metric_worker.wait_final_res() 
                logger.info(f"Initial validation metrics:\n{pformat(val_metrics)}")
                self.logger.log(data=val_metrics, step=self.global_steps)

            if self.config.trainer.val_only:
                logger.info("`val_only` is true. Halting after initial validation.")
                return
        self._run_training_loop()

        if self.progress_bar:
            self.progress_bar.close()
        self.taskgraph_execute_finished = True
        logger.success(f"Rank {self._rank}: DAG execution finished.")

    def _run_training_loop(self):
        """
        The main loop that iterates through training steps and epochs.
        """
        self.total_training_steps = self.dataloader.total_training_steps
        if self.dataloader.num_train_batches <= 0:
            if self._rank == 0:
                logger.warning(f"num_train_batches is {self.dataloader.num_train_batches}. The training loop will be skipped.")
            return

        if self._rank == 0:
            self.progress_bar = tqdm(total=self.total_training_steps, initial=self.global_steps, desc="Training Progress")

        last_val_metrics = None

        # Calculate starting epoch and batches to skip in that epoch for resumption.
        start_epoch = 0
        batches_to_skip = 0
        if self.dataloader.num_train_batches > 0:
            start_epoch = self.global_steps // self.dataloader.num_train_batches
            batches_to_skip = self.global_steps % self.dataloader.num_train_batches

        for epoch in range(start_epoch, self.config.trainer.total_epochs):
            is_embodied = self.config.algorithm.workflow_type == WorkflowType.EMBODIED
            if is_embodied:
                self._cleanup_step_buffers(self.timing_raw)
            for batch_idx in range(self.dataloader.num_train_batches):
                if epoch == start_epoch and batch_idx < batches_to_skip:
                    continue

                if self.global_steps >= self.total_training_steps:
                    logger.info(f"Rank {self._rank}: Reached total training steps. Exiting loop.")
                    if self._rank == 0 and last_val_metrics:
                        logger.info(f"Final validation metrics:\n{pformat(last_val_metrics)}")
                    return
                
                if self.global_steps in self.config.profiler.profile_steps:
                    self._profiler.start(role="e2e", profile_step=self.global_steps)
                    
                ordered_metrics = self._run_training_step(epoch, batch_idx)
                
                if self.global_steps in self.config.profiler.profile_steps:
                    self._profiler.stop()

                if ordered_metrics is None:
                    if self.progress_bar:
                        self.progress_bar.update(1)
                    continue

                self.global_steps += 1

                is_last_step = self.global_steps >= self.total_training_steps

                if self.config.trainer.save_freq > 0 and (is_last_step or self.global_steps % self.config.trainer.save_freq == 0):
                    self.checkpoint_manager.save_checkpoint(self.global_steps)

                if self.config.trainer.test_freq > 0 and (is_last_step or self.global_steps % self.config.trainer.test_freq == 0):
                    self.validator.validate(global_step=self.global_steps)
                    self.metric_worker.wait_submit()
                    dist.barrier(self._gather_group)
                    if self._rank == 0:
                        val_metric = self.metric_worker.wait_final_res()
                        ordered_metrics.update(val_metric)
                        if is_last_step:
                            last_val_metrics = val_metric

                if self.enable_perf:
                    aggregate_and_write_performance_metrics(self._gather_group, self._rank, self.global_steps, self.config, ordered_metrics)
                ordered_metric_dict = format_metrics_by_group(ordered_metrics, DAGConstants.METRIC_GROUP_ORDER)
                log_core_performance_metrics(self._rank, self.enable_perf, ordered_metric_dict, self.global_steps)
                if self._rank == 0:
                    if self.logger:
                        self.logger.log(data=ordered_metric_dict, step=self.global_steps)
                    else:
                        log_metrics_to_console(self._rank, ordered_metric_dict, self.global_steps)

                if self.progress_bar and not (epoch == start_epoch and batch_idx < batches_to_skip):
                    self.progress_bar.update(1)

        if self._rank == 0 and last_val_metrics:
            logger.info(f"Final validation metrics:\n{pformat(last_val_metrics)}")

    def _cleanup_step_buffers(self, timing_raw: dict) -> None:
        """
        Encapsulates the logic for resetting and clearing all step-related buffers.
        This includes the distributed Ray data buffers and the local internal cache.
        This is called at the end of a step, whether it completed successfully or was aborted.
        """
        # Reset the distributed (Ray) buffers for all keys that were used in this step.
        with timer(self.enable_perf, "reset_data_buffer", timing_raw):
            self.reset_data_buffer()
            for ref in self.sample_ref_cache:
                ray.internal.free(ref)
            self.sample_ref_cache = []
        # Clear the local, in-process cache for the next step.
        with timer(self.enable_perf, "reset_intern_data_buffer", timing_raw):
            self.internal_data_cache.clear()

    def _run_training_step(self, epoch: int, batch_idx: int) -> Optional[List[Tuple[str, Any]]]:
        """Executes a single training step by traversing the computational graph."""
        timing_raw, ordered_metrics = self.timing_raw, []

        with timer(self.enable_perf, "step", timing_raw):
            # --- 1. Data Loading ---
            with timer(self.enable_perf, "get_data_from_dataloader", timing_raw):
                batch = preprocess_dataloader(self.dataloader.run(epoch=epoch, is_validation_step=False), self.config.actor_rollout_ref.rollout.n)
            node_queue = self.taskgraph.get_entry_nodes()
            if not node_queue:
                logger.error("Taskgraph has no entry nodes. Cannot start execution.")
                return None
            entry_node_id = node_queue[0].node_id

            # --- 2. Graph Traversal ---
            visited_nodes = set()
            with timer(self.enable_perf, "graph_execution", timing_raw):
                while node_queue:
                    cur_node = node_queue.pop(0)
                    if cur_node.node_id in visited_nodes:
                        continue
                    visited_nodes.add(cur_node.node_id)

                    cur_dp_size, cur_dp_rank, cur_tp_rank, cur_tp_size, cur_pp_rank, cur_pp_size = self._get_node_dp_info(cur_node)
                    logger.debug(f"current node({cur_node.node_id}) dp_size: {cur_dp_size}, dp_rank: {cur_dp_rank}, tp_rank: {cur_tp_rank}, pp_rank: {cur_pp_rank}, pp_size: {cur_pp_size}")

                    # --- 3. Get Input Data ---
                    if cur_node.node_id != entry_node_id:
                        with timer(self.enable_perf, "get_data_from_buffer", timing_raw):
                            batch = self.get_data_from_buffers(key=cur_node.node_id, cur_dp_size=cur_dp_size, cur_dp_rank=cur_dp_rank, timing_raw=timing_raw)
                            if batch is None:
                                if self.config.algorithm.filter_groups.enable:
                                    if cur_node.node_role == NodeRole.ACTOR:
                                        logger.error(f"Rank {self._rank}: Failed to get data for node {cur_node.node_id}. Skipping step.")
                                        return None 
                                else:
                                    logger.error(f"Rank {self._rank}: Failed to get data for node {cur_node.node_id}. Skipping step.")
                                    return None 
                            else:
                                # batch = remove_prefix_from_dataproto(batch, cur_node)
                                logger.debug(f"current node({cur_node.node_id}) get data from databuffer batch size: {batch.size()}")
                    if self.enable_perf:
                        with timer(self.enable_perf, "get_data_from_buffer_barrier", timing_raw):
                            dist.barrier(self._gather_group)
                    # --- 4. Node Execution ---
                    node_name_timer = f"{cur_node.node_id}"
                    with timer(self.enable_perf, node_name_timer, timing_raw):
                        if cur_node.executable and batch is not None:
                            node_kwargs = {"_dag_worker_instance": self}
                            node_kwargs["process_group"] = self._get_node_process_group(cur_node) if cur_node.node_type != NodeType.COMPUTE else None
                            node_kwargs["agent_group"] = self.multi_agent_group[cur_node.agent_group]
                            node_kwargs["cur_tp_rank"] = cur_tp_rank
                            if cur_node.node_role == NodeRole.REWARD:
                                node_kwargs["tp_size"] = cur_tp_size
                            elif cur_node.node_role == NodeRole.ADVANTAGE:
                                node_kwargs["cur_node"] = cur_node

                            if cur_node.agent_options and cur_node.agent_options.train_cycle:
                                cycle_round = self.global_steps // cur_node.agent_options.train_cycle
                                agent_num = len(self.multi_agent_group)
                                if cycle_round % agent_num == cur_node.agent_group:
                                    node_output = cur_node.run(batch=batch,
                                                               config=self.config,
                                                               **node_kwargs)
                                else:
                                    node_output = NodeOutput(batch=batch)
                            else:
                                node_output = cur_node.run(batch=batch,
                                                           config=self.config,
                                                           **node_kwargs)
                        else:
                            logger.warning(f"Node {cur_node.node_id} has no executable. Passing data through.")
                            node_output = NodeOutput(batch=batch)
                    
                    # Check if node returned empty batch (e.g., DAPO insufficient samples)
                    # This triggers re-rollout to collect more data
                    if  node_output.batch is None or (node_output.batch is not None and len(node_output.batch) == 0):
                        logger.warning(
                            f"Rank {self._rank}: Node '{cur_node.node_id}' returned empty batch. "
                        )
                        if not self.config.algorithm.filter_groups.enable:
                            logger.warning(
                                f"Rank {self._rank}: Node '{cur_node.node_id}' returned empty batch. "
                                f"Aborting current step to trigger re-rollout. {node_output.batch is not None and len(node_output.batch) != 0}"
                            )
                            return None
                    
                    if self.enable_perf:        
                        with timer(self.enable_perf, f"{node_name_timer}_barrier", timing_raw):
                            dist.barrier(self._gather_group)
                    if cur_node.node_role == NodeRole.ROLLOUT and self._multi_agent:
                        next_nodes = self.taskgraph.get_downstream_nodes(cur_node.node_id)
                        while next_nodes[0].node_role == NodeRole.ROLLOUT:
                            cur_node = next_nodes[0]
                            next_nodes = self.taskgraph.get_downstream_nodes(cur_node.node_id)

                    # --- 5. Process Output & Get next node ---
                    with timer(self.enable_perf, "graph_output_handling", timing_raw):
                        if node_output.metrics and cur_tp_rank == 0 and cur_pp_rank == 0:
                            self.metric_worker.submit_metric(node_output.metrics, cur_dp_size)
                        if next_nodes := self.taskgraph.get_downstream_nodes(cur_node.node_id):
                            if node_output.batch is not None and len(node_output.batch) != 0:
                                # Currently supports single downstream node, can be extended to a loop.
                                next_node = next_nodes[0]
                                next_dp_size, _, _, _, _, _ = self._get_node_dp_info(next_node)
                                # node_output.batch = add_prefix_to_dataproto(node_output.batch, cur_node)
                                is_current_last_pp_tp_rank0 = (cur_pp_rank == cur_pp_size - 1 and cur_tp_rank == 0)
                                if whether_put_data(self._rank, is_current_last_pp_tp_rank0, next_dp_size, cur_dp_size, cur_node, next_node):
                                    with timer(self.enable_perf, "put_data_to_buffer", timing_raw):
                                        # if self._multi_agent and next_node.node_role == NodeRole.ADVANTAGE:
                                        #     self.multi_agent_put_log(key=next_node.node_id, data=node_output.batch, next_dp_size = next_dp_size, agent_group = next_node.agent_group, timing_raw = timing_raw)
                                        # else:
                                        # have filter, must use databuffer to rebalance
                                        enforce_buffer = (self.config.algorithm.filter_groups.enable) and (cur_node.node_type == NodeType.COMPUTE) and (next_node.node_type == NodeType.MODEL_TRAIN) 
                                        self.put_data_to_buffers(key=next_node.node_id, data=node_output.batch,  source_dp_size=cur_dp_size, dest_dp_size=next_dp_size, enforce_buffer = enforce_buffer, timing_raw=timing_raw)
                        # elif self._multi_agent:
                        #     # last_node add prefix for metrics
                        #     node_output.batch = add_prefix_to_dataproto(node_output.batch, cur_node)                        
                        if self.enable_perf:
                            with timer(self.enable_perf, "put_data_to_buffer_barrier", timing_raw):
                                dist.barrier(self._gather_group)
                        with timer(self.enable_perf, "get_next_node", timing_raw):
                            for n in next_nodes:
                                if n.node_id not in visited_nodes:
                                    node_queue.append(n)

                    with timer(self.enable_perf, "step_barrier", timing_raw):
                        dist.barrier(self._gather_group)

            # --- 6. Final Metrics Collection ---
            self._cleanup_step_buffers(timing_raw)

        ordered_metrics = {}
        if cur_tp_rank == 0 and cur_pp_rank == 0:
            self.metric_worker.compute_local_data_metric(batch, cur_dp_size)
            self.metric_worker.compute_local_throughout_metrics(batch, timing_raw, cur_pp_size * cur_tp_size , cur_dp_size)
            if self._rank == 0:
                # only use rank0 time metrics
                self.metric_worker.compute_local_timing_metrics(batch, timing_raw, 1)  
        timing_raw.clear()
        self.metric_worker.wait_submit()
        dist.barrier(self._gather_group)
        if self._rank == 0:
            metrics = self.metric_worker.wait_final_res()
            ordered_metrics = dict(sorted(metrics.items()))
            ordered_metrics.update({"training/global_step": self.global_steps + 1, "training/epoch": epoch + 1})

        return ordered_metrics

# ==========================================================================================
# Module 2: Graph Node Execution Handlers
# ==========================================================================================

    @DistProfiler.annotate(role="generate")
    def generate_sync_mode(self, agent_group, batch: TensorDict) -> NodeOutput:
        """Sync mode"""
        gen_output = agent_group[NodeRole.ROLLOUT].generate_sequences(batch)
        if "response_mask" not in batch:
            gen_output["response_mask"] = compute_response_mask(gen_output)
        batch = batch.update(gen_output)
        return NodeOutput(batch=batch, metrics=gen_output["metrics"])

    @DistProfiler.annotate(role="generate")
    def generate_async_mode(self, batch: TensorDict) -> NodeOutput:
        """Async mode"""
        if self._async_rollout_manager is not None:
            loop = asyncio.get_event_loop()
            gen_output = loop.run_until_complete(self._async_rollout_manager.generate_sequences(batch))
            metrics = gen_output["metrics"]
            if "response_mask" not in batch:
                batch["response_mask"] = compute_response_mask(batch)
            return NodeOutput(batch=batch, metrics=metrics)
        return NodeOutput(batch=batch, metrics={})

    @DistProfiler.annotate(role="generate")
    def generate_multi_agent_mode(self, config, batch: TensorDict) -> NodeOutput:
        """Generates sequences for a training batch using the multi-agent rollout model."""
        gen_batch = prepare_generation_batch(batch)
        if config.actor_rollout_ref.rollout.agent.rewards_with_env and "reward_model" in batch.non_tensor_batch:
            gen_batch.non_tensor_batch["reward_model"] = batch.non_tensor_batch["reward_model"]
        assert config.actor_rollout_ref.rollout.name == 'sglang'
        gen_output = self.multi_agent_loop.generate_sequence(gen_batch)
        if gen_output:
            metrics = gen_output.meta_info.get("metrics", {})
            # gen_output.meta_info = {}
            # batch.non_tensor_batch["uid"] = np.array([str(uuid.uuid4()) for _ in range(len(batch.batch))])
            # batch = batch.repeat(config.actor_rollout_ref.rollout.n, interleave=True).union(gen_output)
            # if "response_mask" not in batch.batch:
            #     batch.batch["response_mask"] = compute_response_mask(batch)
            return NodeOutput(batch=gen_output, metrics=metrics)
        return NodeOutput(batch=batch, metrics={})

    @DistProfiler.annotate(role="generate")
    def generate_embodied_mode(self, agent_group, batch: TensorDict, **kwargs) -> NodeOutput:
        """
        Generates embodied episodes for training.
        
        This method follows the same pattern as _generate_for_embodied_validation in validation_mixin,
        but configured for training mode (do_sample=True, validate=False).
        
        For embodied tasks, the batch contains task metadata (task_id, trial_id, etc.) from the dataloader.
        The rollout worker interacts with the environment and generates all required data
        (input_ids, pixel_values, responses, etc.) during environment rollout.
        
        Unlike text generation, we do NOT call _prepare_generation_batch because:
        1. The input batch doesn't have text-generation keys (input_ids, attention_mask, etc.)
        2. These keys will be generated by the embodied rollout worker during env interaction
        """
        from loguru import logger
        
        rollout_worker = agent_group[NodeRole.ROLLOUT]
        
        # Set meta_info for embodied training
        batch["eos_token_id"] = NonTensorData(self.validate_tokenizer.eos_token_id if self.validate_tokenizer else None)
        batch["n_samples"] = NonTensorData(self.config.actor_rollout_ref.rollout.n)
        batch["pad_token_id"] = NonTensorData(self.validate_tokenizer.pad_token_id if self.validate_tokenizer else None)        
        logger.info(
            f"[Embodied Validation] Batch variables: "
            f"{batch.batch_size[0]}, "
            f"eos_token_id={batch['eos_token_id']}, "
            f"pad_token_id={batch['pad_token_id']}, "
            f"n_samples={batch['n_samples']}, "
        )
        # Generate embodied episodes
        gen_output = rollout_worker.generate_sequences(batch)
        metrics = gen_output["metrics"]
        batch.update(gen_output)
        # Add unique IDs for tracking
        # Compute response mask if not already present
        if "response_mask" not in batch:
            batch["response_mask"] = compute_response_mask(batch)
        
        return NodeOutput(batch=batch, metrics=metrics)
    
    
    
    @DistProfiler.annotate(role="generate")
    def generate(self, config, batch: TensorDict, **kwargs) -> NodeOutput:
        """Generates sequences for a training batch using the rollout model."""
        # Check if this is embodied mode
        agent_group = kwargs.pop("agent_group")
        is_embodied = self.config.actor_rollout_ref.model.model_type == "embodied"
        
        if is_embodied:
            # Use dedicated embodied generation path (mirrors validation logic)
            return self.generate_embodied_mode(agent_group, batch, **kwargs)
        if self._multi_agent is False:
            if self.rollout_mode == 'sync':
                return self.generate_sync_mode(agent_group, batch)
            else:
                return self.generate_async_mode(batch)
        else:
            return self.generate_multi_agent_mode(config, batch)

    @DistProfiler.annotate(role="compute_reward")
    def compute_reward(self, config, batch: TensorDict, **kwargs) -> NodeOutput:
        """Calculates rewards for a batch of generated sequences."""
        
        if not self.check_mode() and kwargs["cur_tp_rank"] != 0:
            return NodeOutput(batch=batch, metrics={})
        
        tp_size = kwargs.pop("tp_size")
        if "token_level_rewards" in batch and batch["token_level_rewards"].numel() > 0:
            return NodeOutput(batch=batch, metrics={})
        batch["global_token_num"] = NonTensorData((torch.sum(batch["attention_mask"], dim=-1) // tp_size).tolist())

        reward_tensor, extra_infos = compute_reward(batch, self.reward_fn)
        batch["token_level_scores"] = reward_tensor

        if extra_infos:
            batch.update({k: np.array(v) for k, v in extra_infos.items()}, inplace=True)

        metrics = {}
        if config.algorithm.use_kl_in_reward:
            kl_ctrl_in_reward = core_algos.get_kl_controller(config.algorithm.kl_ctrl)
            batch, kl_metrics = apply_kl_penalty(batch, kl_ctrl_in_reward, config.algorithm.kl_penalty)
            metrics.update(kl_metrics)
        else:
            batch["token_level_rewards"] = batch["token_level_scores"]
        return NodeOutput(batch=batch, metrics=metrics)

    
    @DistProfiler.annotate(role="compute_old_log_prob")
    def compute_old_log_prob(self, config, batch: TensorDict, **kwargs) -> NodeOutput:
        """Computes log probabilities from the actor model before the policy update."""
        process_group = kwargs.pop("process_group")
        agent_group = kwargs.pop("agent_group")
        if "global_token_num" not in batch:
            # in multi-agent, agentA may don't have reward node
            # insert some info needed
            batch["global_token_num"] = NonTensorData(torch.sum(batch["attention_mask"], dim=-1).tolist())
        processed_data = agent_group[NodeRole.ACTOR].compute_log_prob(batch)
        local_metrics = processed_data["metrics"]  if "metrics" in processed_data else {}
        if "entropys" in processed_data:
            entropy = agg_loss(processed_data["entropys"], processed_data["response_mask"].to("cpu"), config.actor_rollout_ref.actor.loss_agg_mode)
            local_metrics["actor/entropy_loss"] = entropy.item()

        processed_data.pop("metrics", None)
        processed_data.pop("entropys", None)

        return NodeOutput(batch=processed_data, metrics=local_metrics)

    @DistProfiler.annotate(role="compute_ref_log_prob")
    def compute_ref_log_prob(self, config, batch: TensorDict, **kwargs) -> NodeOutput:
        """Computes log probabilities from the frozen reference model."""
        agent_group = kwargs.pop("agent_group")
        processed_data = agent_group[NodeRole.REFERENCE].compute_ref_log_prob(batch)
        metrics = processed_data["metrics"]
        return NodeOutput(batch=processed_data, metrics=metrics)

    @DistProfiler.annotate(role="compute_value")
    def compute_value(self, config, batch: TensorDict, **kwargs) -> NodeOutput:
        """Computes value estimates from the critic model."""
        agent_group = kwargs.pop("agent_group")
        processed_data = agent_group[NodeRole.CRITIC].compute_values(batch)
        return NodeOutput(batch=processed_data)

    @DistProfiler.annotate(role="compute_advantage")
    def compute_multi_agent_advantage(self, config, batch: TensorDict, **kwargs) -> NodeOutput:
        adv_config = config.algorithm
        rollout_config = config.actor_rollout_ref.rollout
        cur_node = kwargs["cur_node"]
        if "token_level_rewards" not in batch.batch :
            # make sure rewards of angentB has been compute
            # GAE_MARFT adv need make sure only last agent has adv node
            if depend_nodes := self.taskgraph.get_dependencies(cur_node.node_id):
                depend_node = depend_nodes[0]
                if adv_config.share_reward_in_agent:
                    batch.batch["token_level_rewards"] = batch.batch[f"agent_group_{depend_node.agent_group}_token_level_rewards"].clone()
                else:
                    batch.batch["token_level_rewards"] = torch.zeros_like(batch.batch[f"agent_group_{depend_node.agent_group}_token_level_rewards"])
                batch.batch["token_level_scores"] = batch.batch[f"agent_group_{depend_node.agent_group}_token_level_scores"].clone()
            else:
                raise RuntimeError(f"cur_node {cur_node.node_id} have no rewards with can't find it's dependencies reward")
        if adv_config.adv_estimator == AdvantageEstimator.GAE_MARFT:
            # make sure adv node define in last agent node
            cur_agent_id = len(self.multi_agent_group) - 1
            agent_groups_ids = list(range(cur_agent_id))
            kwargs["agent_group_ids"] = agent_groups_ids
            # pre_agent may have no reward token
            for agent_id in reversed(agent_groups_ids):
                key_prefix = f"agent_group_{agent_id}_token_level_rewards"
                if key_prefix not in batch.batch:
                    pre_key_prefix = f"agent_group_{agent_id + 1}_token_level_rewards" if agent_id != cur_agent_id -1 else "token_level_rewards"
                    if adv_config.share_reward_in_agent:
                        batch.batch[key_prefix] = batch.batch[pre_key_prefix].clone()
                    else:
                        batch.batch[key_prefix] = torch.zeros_like(batch.batch[pre_key_prefix])
                batch.batch[f"agent_group_{agent_id}_token_level_scores"] = batch.batch[key_prefix].clone()

        return NodeOutput(
            batch=compute_advantage(
                batch,
                adv_estimator=adv_config.adv_estimator,
                gamma=adv_config.gamma,
                lam=adv_config.lam,
                num_repeat=rollout_config.n,
                norm_adv_by_std_in_grpo=adv_config.norm_adv_by_std_in_grpo,
                weight_factor_in_cpgd=adv_config.weight_factor_in_cpgd,
                multi_turn=rollout_config.multi_turn.enable,
                **kwargs
            )
        )

    @DistProfiler.annotate(role="compute_advantage")
    def compute_advantage(self, config, batch: TensorDict, **kwargs) -> NodeOutput:
        """Computes advantages and returns for PPO using GAE."""
        
        if not self.check_mode() and kwargs["cur_tp_rank"] != 0:
            return NodeOutput(batch=batch, metrics={})
        
        if self._multi_agent:
            return self.compute_multi_agent_advantage(config, batch, **kwargs)
        algo_config = config.algorithm
        return NodeOutput(
            batch=compute_advantage(
                batch,
                adv_estimator=algo_config.adv_estimator,
                gamma=algo_config.gamma,
                lam=algo_config.lam,
                norm_adv_by_std_in_grpo=algo_config.norm_adv_by_std_in_grpo,
                weight_factor_in_cpgd=algo_config.weight_factor_in_cpgd,
                **kwargs
            )
        )

    @DistProfiler.annotate(role="train_critic")
    def train_critic(self, config, batch: TensorDict, **kwargs) -> NodeOutput:
        """Performs a single training step on the critic model."""
        agent_group = kwargs.pop("agent_group")
        process_group = kwargs.pop("process_group")
        processed_data = agent_group[NodeRole.CRITIC].update_critic(batch)
        return NodeOutput(batch=processed_data, metrics=processed_data["metrics"])

    @DistProfiler.annotate(role="train_actor")
    def train_actor(self, config, batch: TensorDict, **kwargs) -> NodeOutput:
        """Performs a single training step on the actor (policy) model."""
        process_group = kwargs.pop("process_group")
        agent_group = kwargs.pop("agent_group")
        global_steps = batch["global_steps"] if "global_steps" in batch else 0
        if config.trainer.critic_warmup > global_steps:
            return NodeOutput(batch=batch)  # Skip actor update during critic warmup
        batch["multi_turn"] = NonTensorData(self.config.actor_rollout_ref.rollout.multi_turn.enable)
        processed_data = agent_group[NodeRole.ACTOR].update_actor(batch)
        return NodeOutput(batch=processed_data, metrics=processed_data["metrics"])


# ==========================================================================================
# Module 3: Worker and Environment Initialization
# ==========================================================================================

    def _initialize_worker(self):
        """Orchestrates the ordered initialization of all worker components."""
        self._rank = get_and_validate_rank()
        self.taskgraph = get_taskgraph_for_rank(self._rank, self.taskgraph_mapping)

        self._setup_distributed_environment()
        self._setup_tokenizers()
        self._setup_dataloader()
        self._setup_reward_managers()
        self._setup_role_worker_mapping()
        self._initialize_node_workers()
        self._profiler = DistProfiler(rank=self._rank, config=self.config.profiler)

        # Initialize CheckpointManager - Note: will be fully initialized after workers are created
        self.checkpoint_manager = None

        # Initialize Validator - Note: will be initialized in init_graph() after all workers are ready
        self.validator = None

        # Initialize MetricsCollector - Note: will be initialized in init_graph() after all dependencies are ready
        self.metrics_collector = None

        if self._rank == 0:
            logger.info("Rank 0: Initializing tracking logger...")
            from siirl.utils.logger.tracking import Tracking

            self.logger = Tracking(
                project_name=self.config.trainer.project_name,
                experiment_name=self.config.trainer.experiment_name,
                default_backend=self.config.trainer.logger,
                config=self.config.to_dict(),
            )
            if self.enable_perf:
                logger.warning("Performance tracking is enabled. This may impact training speed.")

    def _setup_distributed_environment(self):
        """Initializes the default process group and all required subgroups."""

        if not dist.is_initialized():
            backend = (
                f"{get_nccl_backend()}"
                if self.world_size >= self.config.dag.backend_threshold
                else f"cpu:gloo,{get_device_name()}:{get_nccl_backend()}"
            )
            logger.info(
                f"Rank {self._rank}: Initializing world size {self.world_size} default process group with '{backend}' "
                f"backend."
            )
            dist.init_process_group(backend=backend)

        if device_name == "npu":
            # For NPU, metrics aggregation requires the hccl backend for device-to-device communication.
            # This group is created regardless of world size for NPU environments.
            gather_backend = get_nccl_backend()
            self._gather_group = dist.new_group(backend=gather_backend)
        else:
            # For GPU, the original logic is preserved for backward compatibility.
            # The gather group is only created if world_size < backend_threshold.
            self._gather_group = dist.new_group(
                backend="gloo") if self.world_size < self.config.dag.backend_threshold else None

        group_specs = self.process_group_manager.get_all_specs()
        if not group_specs:
            logger.warning("No process group specifications found in ProcessGroupManager.")
            return

        #Builds all process groups defined in the ProcessGroupManager.
        for name, spec in group_specs.items():
            if not isinstance(spec, dict) or not (ranks := spec.get("ranks")):
                logger.warning(f"Skipping group '{name}' due to invalid spec or missing 'ranks'.")
                continue
            self.process_groups[name] = dist.new_group(ranks=ranks)
        logger.debug(f"Rank {self._rank}: Created {len(self.process_groups)} custom process groups.")

        self.inference_group_name_set = self.process_group_manager.get_process_group_for_node_type_in_subgraph(
            self.taskgraph.graph_id, NodeType.MODEL_INFERENCE.value
        )
        self.train_group_name_set = self.process_group_manager.get_process_group_for_node_type_in_subgraph(
            self.taskgraph.graph_id, NodeType.MODEL_TRAIN.value
        )

        # Ensure all ranks have finished group creation before proceeding.
        dist.barrier(self._gather_group)
        logger.info(f"Rank {self._rank}: Distributed environment setup complete.")

    def _setup_tokenizers(self):
        """Initializes and caches tokenizers for all models in the task graph."""
        model_nodes = [
            node
            for node in self.taskgraph.nodes.values()
            if node.node_type in [NodeType.MODEL_TRAIN, NodeType.MODEL_INFERENCE]
        ]
        if not model_nodes:
            logger.warning("No model nodes found in the task graph. Tokenizer setup will be skipped.")
            return

        for node in model_nodes:
            agent_key = f"group_key_{node.agent_group}"
            if agent_key not in self.tokenizer_mapping:
                # Add robust check for missing configuration.
                intern_config = node.config.get(DAGConstants.INTERN_CONFIG)
                if not intern_config or not (model_dict := getattr(intern_config, "model", None)):
                    logger.warning(f"Node {node.node_id} is missing model config. Skipping tokenizer setup for it.")
                    continue

                tokenizer_module = load_tokenizer(model_args=model_dict)
                if tokenizer := tokenizer_module.get("tokenizer"):
                    tokenizer.padding_side = "left"  # Required for most causal LM generation
                self.tokenizer_mapping[agent_key] = tokenizer_module
        logger.info(f"Rank {self._rank}: Initialized {len(self.tokenizer_mapping)} tokenizer(s).")

    def _setup_dataloader(self):
        """Initializes the data loader for training and validation."""
        rollout_nodes = [n for n in self.taskgraph.nodes.values() if n.node_type == NodeType.MODEL_INFERENCE]
        if not rollout_nodes:
            raise ValueError("At least one MODEL_INFERENCE node is required for dataloader setup.")
        self.first_rollout_node = rollout_nodes[0]

        pg_assignment = self.process_group_manager.get_node_assignment(self.first_rollout_node.node_id)
        if not (process_group_name := pg_assignment.get("process_group_name")):
            raise ValueError(
                f"Process group name not found for the first rollout node {self.first_rollout_node.node_id}."
            )

        self.dataloader_process_group = self.process_groups.get(process_group_name)
        if self.dataloader_process_group is None:
            raise ValueError(f"Could not find process group '{process_group_name}' in the created groups.")

        self.dataloader_tensor_model_parallel_size = self.first_rollout_node.config[
            DAGConstants.INTERN_CONFIG
        ].rollout.tensor_model_parallel_size

        self.dataloader = DataLoaderNode(
            node_id="dataloader",
            global_config=self.config,
            config={
                "group_world_size": dist.get_world_size(self.dataloader_process_group),
                "group_rank": dist.get_rank(self.dataloader_process_group),
                "group_parallel_size": self.dataloader_tensor_model_parallel_size,
                "num_loader_workers": self.config.data.num_loader_workers,
                "auto_repeat": self.config.data.auto_repeat,
            },
        )
        logger.info(f"Rank {self._rank}: DataLoader initialized with {self.dataloader.total_training_steps} total training steps.")

    def _setup_reward_managers(self):
        """Initializes reward managers for training and validation."""
        self.validate_tokenizer = next(iter(self.tokenizer_mapping.values()), {}).get("tokenizer")
        if not self.validate_tokenizer:
            logger.warning("No tokenizer loaded; reward functions might fail or use a default one.")

        self.reward_fn = create_reward_manager(
            self.config,
            self.validate_tokenizer,
            num_examine=0,
            max_resp_len=self.config.data.max_response_length,
            overlong_buffer_cfg=self.config.reward_model.overlong_buffer,
            **self.config.reward_model.reward_kwargs,
        )
        logger.info(f"Rank {self._rank}: Reward managers initialized.")

    def _setup_role_worker_mapping(self):
        """Creates a mapping from NodeRole to the corresponding Worker implementation class."""
        self.role_worker_mapping: Dict[NodeRole, Type[Worker]] = {}
        # Actor/Ref/Rollout/Critic workers
        actor_strategy = self.config.actor_rollout_ref.actor.strategy
        self.role_worker_mapping.update(get_worker_classes(self.config, actor_strategy))

        # Reward model worker (if enabled)
        if self.config.reward_model.enable:
            reward_strategy = self.config.reward_model.strategy
            reward_workers = get_worker_classes(self.config, reward_strategy)
            if NodeRole.REWARD in reward_workers:
                self.role_worker_mapping[NodeRole.REWARD] = reward_workers[NodeRole.REWARD]
            else:
                logger.warning(
                    f"Reward model is enabled, but no worker found for role REWARD with strategy {reward_strategy}."
                )

        log_role_worker_mapping(self.role_worker_mapping)

    def _initialize_node_workers(self):
        """Instantiates worker objects for all nodes in the task graph."""
        for node in self.taskgraph.nodes.values():
            if not should_create_worker(self.role_worker_mapping, node):
                continue

            worker_cls = self.role_worker_mapping.get(node.node_role)
            if not worker_cls:
                logger.warning(f"No worker class found for role {node.node_role.name}. Skipping node {node.node_id}.")
                continue

            node_worker_key = generate_node_worker_key(node)
            if node_worker_key in self.workers:
                continue

            try:
                node_process_group = self._get_node_process_group(node)
                config = node.config.get(DAGConstants.INTERN_CONFIG)
                if hasattr(config, "actor") and hasattr(config.actor, "optim"):
                    config.actor.optim.total_training_steps = self.dataloader.total_training_steps
                elif hasattr(config, "optim"):
                    config.optim.total_training_steps = self.dataloader.total_training_steps
                worker_args = {"config": config, "process_group": node_process_group}

                # For separated workers (Megatron backend), no role parameter is needed
                # Only legacy ActorRolloutRefWorker needs the role parameter
                if hasattr(worker_cls, '__name__') and 'ActorRolloutRefWorker' in worker_cls.__name__:
                    if node.node_role in DAGConstants.WORKER_ROLE_MAPPING:
                        worker_args["role"] = DAGConstants.WORKER_ROLE_MAPPING[node.node_role]
                if node.agent_options and node.agent_options.share_instance:
                    # cur agent share same critic with target agent
                    self.multi_agent_group[node.agent_group][node.node_role] = self.multi_agent_group[node.agent_options.share_instance][node.node_role]
                else:
                    worker_instance = worker_cls(**worker_args)
                    self.workers[node_worker_key] = worker_instance
                    self.multi_agent_group[node.agent_group][node.node_role] = worker_instance
                    self.agent_group_process_group[node.agent_group][node.node_role] = node_process_group
                    logger.success(
                        f"Rank {self._rank}: Successfully created worker '{worker_cls.__name__}' for node: {node.node_id}"
                    )

            except Exception as e:
                #  Explicitly log the failing node and worker class, then re-raise
                # the exception to prevent silent failures.
                logger.error(
                    f"Failed to create worker for node {node.node_id} with class {worker_cls.__name__}.", exc_info=True
                )
                raise RuntimeError(f"Worker instantiation failed for node {node.node_id}") from e

        if len(self.multi_agent_group) > 1:
            self._multi_agent = True

    def init_graph(self):
        """
        Initializes the computation graph by loading models and restoring checkpoint state.

        Executed after _initialize_worker() across all workers via Ray remote call.
        This method include:
        (1) model weight loading,
        (2) weight sharding_manager setup,
        (3) async/multi-agent init,
        (4) validator init,
        (5) metrics collector init,
        (6) checkpoint restoration
        """

        self._load_model_weights()

        self._setup_sharding_manager()

        self._setup_async_rollout()

        self._setup_multi_agent_loop()

        self._init_validator()

        self._init_metrics_collector()

        self._init_checkpoint_manager()
        self.global_steps = self.checkpoint_manager.load_checkpoint()

        dist.barrier(self._gather_group)

    def _load_model_weights(self):
        """Loads model weights to GPU for all node workers."""
        logger.info("Loading model weights for all worker nodes...")
        initialized_workers = set()

        for node in self.taskgraph.nodes.values():
            if not should_create_worker(self.role_worker_mapping, node):
                continue

            worker_key = generate_node_worker_key(node)
            if worker_key in initialized_workers:
                continue

            node_worker = self.workers[worker_key]
            if not isinstance(node_worker, Worker):
                raise TypeError(f"Invalid worker type for node {node.node_id}: {type(node_worker).__name__}")

            node_worker.init_model()
            initialized_workers.add(worker_key)

        logger.success("All model weights loaded successfully.")

    def _setup_sharding_manager(self):
        """Sets up sharding managers for actor-rollout weight synchronization."""
        logger.info(f"Setting up weight sharing infrastructure ({self.config.actor_rollout_ref.rollout.name})...")

        for agent_group, worker_dict in self.multi_agent_group.items():
            if NodeRole.ACTOR in worker_dict and NodeRole.ROLLOUT in worker_dict:
                try:
                    setup_sharding_manager(
                        self.config,
                        self.agent_group_process_group,
                        agent_group,
                        worker_dict
                    )
                except Exception as e:
                    logger.error(f"Failed to set up sharding manager for agent group {agent_group}: {e}", exc_info=True)
                    raise

        logger.success("Weight sharing infrastructure initialized.")

    def _setup_async_rollout(self):
        """Initializes async rollout server if configured."""
        if self.config.actor_rollout_ref.rollout.mode != "async":
            return

        logger.info("Initializing async rollout server...")
        for node in self.taskgraph.nodes.values():
            if node.node_role == NodeRole.ROLLOUT:
                self.rollout_mode = "async"
                node_worker = self.workers[generate_node_worker_key(node)]
                self.zmq_address = node_worker.get_zeromq_address()
                self.init_async_server(node=node, node_worker=node_worker)

        logger.success("Async rollout server initialized.")

    def _setup_multi_agent_loop(self):
        """Initializes multi-agent loop if in multi-agent mode."""
        if not self._multi_agent:
            return

        logger.info("Initializing multi-agent loop...")
        from siirl.execution.rollout_flow.multi_agent.multiagent_generate import MultiAgentLoop

        self.multi_agent_loop = MultiAgentLoop(
            self,
            config=self.config.actor_rollout_ref,
            node_workers=self.workers,
            local_dag=self.taskgraph,
            databuffer=self.data_buffers,
            placement_mode='colocate'
        )

        logger.success("Multi-agent loop initialized.")

    def _init_validator(self):
        """Initializes validator for validation workflow."""
        logger.info("Initializing validator...")
        from siirl.dag_worker.validator import Validator

        self.validator = Validator(
            config=self.config,
            dataloader=self.dataloader,
            validate_tokenizer=self.validate_tokenizer,
            multi_agent_group=self.multi_agent_group,
            rollout_mode=self.rollout_mode,
            async_rollout_manager=self._async_rollout_manager,
            multi_agent_loop=getattr(self, 'multi_agent_loop', None),
            multi_agent=self._multi_agent,
            rank=self._rank,
            world_size=self.world_size,
            gather_group=self._gather_group,
            first_rollout_node=self.first_rollout_node,
            get_node_dp_info_fn=self._get_node_dp_info,
            enable_perf=self.enable_perf,
            metric_worker=self.metric_worker
        )
        logger.success("Validator initialized.")

    def _init_metrics_collector(self):
        """Initializes metrics collector for training metrics aggregation."""
        logger.info("Initializing metrics collector...")
        # from siirl.dag_worker.metrics_collector import MetricsCollector
        # self.metric_worker.init()
        # self.metrics_collector = MetricsCollector(
        #     rank=self._rank,
        #     world_size=self.world_size,
        #     gather_group=self._gather_group,
        #     taskgraph=self.taskgraph,
        #     first_rollout_node=self.first_rollout_node,
        #     get_node_dp_info_fn=self._get_node_dp_info,
        #     multi_agent=self._multi_agent,
        #     enable_perf=self.enable_perf,
        # )
        # logger.success("Metrics collector initialized.")

    def _init_checkpoint_manager(self):
        """Initializes checkpoint manager for saving/loading training state."""
        logger.info("Initializing checkpoint manager...")
        self.checkpoint_manager = CheckpointManager(
            config=self.config,
            rank=self._rank,
            gather_group=self._gather_group,
            workers=self.workers,
            taskgraph=self.taskgraph,
            dataloader=self.dataloader,
            first_rollout_node=self.first_rollout_node,
            get_node_dp_info_fn=self._get_node_dp_info
        )

    def init_async_server(self, node:Node, node_worker):
        #gather zmq_address to rank_0
        _, dp_rank, tp_rank, tp_size, *_ = self._get_node_dp_info(node)
        addr_len = len(self.zmq_address)
        encoded_addr = torch.tensor([ord(c) for c in self.zmq_address], dtype=torch.uint8,
                                device=torch.device("cuda" if torch.cuda.is_available() else "cpu"))
        zmq_addresses = []
        if tp_rank == 0:
            group_addrs = torch.zeros((tp_size, addr_len), dtype=torch.uint8, device=encoded_addr.device)
            group_addrs[0] = encoded_addr
            for i in range(1, tp_size):
                src_rank = dp_rank * tp_size + i
                dist.recv(group_addrs[i], src=src_rank)
            for i in range(tp_size):
                addr_str = ''.join([chr(c.item()) for c in group_addrs[i]])
                zmq_addresses.append(addr_str)
        else:
            dist.send(encoded_addr, dst=dp_rank * tp_size)
        if tp_rank == 0:
            self._async_rollout_manager = AgentLoopManager(node.config["intern_config"], dp_rank, os.environ['WG_PREFIX'], node_worker.rollout, zmq_addresses)

# ==========================================================================================
# Module 4: Utilities
# ==========================================================================================
    def put_data_to_buffers(
        self, key: str,
        data: TensorDict,
        source_dp_size:int,
        dest_dp_size: int,
        enforce_buffer: bool,
        timing_raw: Dict[str, float]
    ):
        """
        Puts data into the DataCoordinator by converting it into individual Samples.
        The data is tagged with a 'key' to be retrieved by the correct downstream node.
        """
        try:
            batch_size = len(data) if data is not None else 0

            if source_dp_size == dest_dp_size and not enforce_buffer:
                with timer(self.enable_perf, f"put_intern_data_{key}", timing_raw):
                    self.internal_data_cache[key] = data
            else:
                samples = Dict2Samples(data)
                if not samples:
                    logger.warning(f"Rank {self._rank}: TensorDict for key '{key}' converted to 0 samples. Nothing to put.")
                    return

                with timer(self.enable_perf, f"put_samples_to_coordinator_{key}", timing_raw):
                    sample_infos = []
                    for sample in samples:
                        # Convert uid to string (handle tensor uid from postprocess_sampling)
                        uid_val = getattr(sample, 'uid', uuid.uuid4().int)
                        if isinstance(uid_val, torch.Tensor):
                            uid_str = str(int(uid_val.item()))
                        else:
                            uid_str = str(uid_val)
                        
                        sample_infos.append(SampleInfo(
                            sum_tokens=getattr(sample, 'sum_tokens', int(sample.attention_mask.sum())),
                            prompt_length=getattr(sample, 'prompt_length', 0),
                            response_length=getattr(sample, 'response_length', 0),
                            uid=uid_str,
                            dict_info={
                                'key': key,
                                'source_dp_size': source_dp_size  # Store source DP size
                            }
                        ))
                    
                    # Although ray.put is called multiple times, it is more efficient than remote actor calls.
                    # This is the main source of the remaining overhead, but it is necessary
                    # to maintain sample-level traceability in the DataCoordinator.
                    with timer(self.enable_perf, f"ray_put_samples_{key}", timing_raw):
                        sample_refs = [ray.put(sample) for sample in samples]
                    self.sample_ref_cache.extend(sample_refs)
                    try:
                        loop = asyncio.get_event_loop()
                    except RuntimeError:
                        loop = asyncio.new_event_loop()
                        asyncio.set_event_loop(loop)
                    
                    # Get the current worker's node ID to pass to the DataCoordinator
                    # This is necessary because when the DataCoordinator receives a remote call,
                    # ray.get_runtime_context().get_node_id() returns the DataCoordinator's node_id,
                    # not the caller's node_id
                    caller_node_id = ray.get_runtime_context().get_node_id()
                    
                    put_future = self.data_coordinator.put_batch.remote(sample_infos, sample_refs, caller_node_id)
                    loop.run_until_complete(put_future)
                    logger.info(f"Rank {self._rank}: ✅ Successfully PUT {len(samples)} samples to DataCoordinator for key '{key}' (source_dp={source_dp_size}, dest_dp={dest_dp_size})")

        except Exception as e:
            logger.error(f"Rank {self._rank}: Unexpected error in put_data_to_buffers for key '{key}': {e}", exc_info=True)
            raise

    def get_data_from_buffers(
        self,
        key: str,
        cur_dp_size: int,
        cur_dp_rank: int,
        timing_raw: Dict[str, float]
    ) -> Optional[TensorDict]:
        """
        Gets data from the DataCoordinator by filtering for a specific key,
        then collates the resulting Samples back into a single TensorDict.
        
        Args:
            key: The key to filter samples
            cur_dp_size: Current node's DP size
            cur_dp_rank: Current worker's DP rank
            timing_raw: Timing dict for performance tracking
        """
        with timer(self.enable_perf, f"get_intern_data_{key}", timing_raw):
            if key in self.internal_data_cache:
                cached_data = self.internal_data_cache.pop(key)
                return cached_data
        def key_filter(sample_info: SampleInfo) -> bool:
            return sample_info.dict_info.get('key') == key

        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        with timer(self.enable_perf, f"get_samples_from_coordinator_{key}", timing_raw):
            try:
                rollout_n = self.config.actor_rollout_ref.rollout.n if hasattr(self.config, 'actor_rollout_ref') else 1
            except (AttributeError, KeyError):
                rollout_n = 1
            
            if rollout_n is None or rollout_n < 1:
                rollout_n = 1
            
            adjusted_batch_size = int(self.config.data.train_batch_size * rollout_n / cur_dp_size)
            
            logger.info(
                f"Rank {self._rank}: Requesting from DataCoordinator: "
                f"key='{key}', cur_dp={cur_dp_size}, "
                f"adjusted_batch_size={adjusted_batch_size} (train_bs={self.config.data.train_batch_size} * rollout_n={rollout_n} / cur_dp={cur_dp_size})"
            )
            
            # Use filter_plugin to get only samples with matching key
            # Use balance_partitions to optimize sample distribution by length
            sample_refs = loop.run_until_complete(
                self.data_coordinator.get_batch.remote(
                    adjusted_batch_size,
                    cur_dp_rank,
                    filter_plugin=key_filter,
                    balance_partitions=cur_dp_size
                )
            )

        if not sample_refs:
            logger.warning(f"Rank {self._rank}: ❌ DataCoordinator returned EMPTY list for key '{key}' (adjusted_batch_size={adjusted_batch_size})")
            return None

        logger.info(f"Rank {self._rank}: ✅ Retrieved {len(sample_refs)} sample references from DataCoordinator for key '{key}'")

        with timer(self.enable_perf, f"ray_get_samples_{key}", timing_raw):
            samples = ray.get(sample_refs)

        with timer(self.enable_perf, f"collate_samples_{key}", timing_raw):
            # Collate the list of Sample objects back into a single TensorDict
            tensordict = Samples2Dict(samples)

        return tensordict

    def reset_data_buffer(self):
        """
        DEPRECATED with DataCoordinator. The get calls are now consuming.
        This can be a no-op, but for safety, we could implement a clear if needed.
        For now, it does nothing as intended.
        """
        logger.debug("`reset_data_buffer` is a no-op with the new DataCoordinator model as gets are consuming.")
        if self._rank == 0:
            self.data_coordinator.reset_cache.remote()

    def _get_node_process_group(self, node: Node) -> ProcessGroup:
        """Retrieves the PyTorch ProcessGroup assigned to a specific graph node."""
        assignment = self.process_group_manager.get_node_assignment(node.node_id)
        if not (assignment and (name := assignment.get("process_group_name"))):
            raise ValueError(f"Process group assignment or name not found for node {node.node_id}.")

        pg = self.process_groups.get(name)
        if pg is None:
            raise ValueError(f"Process group '{name}' for node {node.node_id} was not created or found.")
        return pg

    def _get_node_dp_info(self, node: Node) -> tuple[int, int, int, int, int, int]:
        """
        Calculates Data Parallel (DP), Tensor Parallel (TP), and Pipeline Parallel (PP) info for a node.

        Returns:
            tuple: (dp_size, dp_rank, tp_rank, tp_size, pp_rank, pp_size)
        """
        reference_node = node
        if node.node_type == NodeType.COMPUTE:
            # If the node is a COMPUTE type, find its true data source ancestor.
            ancestor = find_first_non_compute_ancestor(self.taskgraph, node.node_id)
            if ancestor:
                reference_node = ancestor
            else:
                # If no non-COMPUTE ancestor is found, it's a critical error.
                raise RuntimeError(f"Could not find any non-COMPUTE ancestor for COMPUTE node '{node.node_id}'. Please check your DAG graph configuration.")

        if reference_node.node_type == NodeType.COMPUTE:
            group_world_size = self.config.trainer.n_gpus_per_node * self.config.trainer.nnodes
            group_rank = dist.get_rank()
        else:
            process_group = self._get_node_process_group(reference_node)
            group_world_size = dist.get_world_size(process_group)
            group_rank = dist.get_rank(process_group)

        # Get parallelism configuration based on backend strategy
        tp_size, pp_size = get_parallelism_config(reference_node)

        # Calculate total parallel size (TP * PP)
        total_parallel_size = tp_size * pp_size

        if group_world_size % total_parallel_size != 0:
            raise ValueError(f"Configuration error for node {node.node_id}: Group world size ({group_world_size}) is not divisible by total parallel size (TP={tp_size} * PP={pp_size} = {total_parallel_size}). Check your parallel configuration.")

        dp_size = group_world_size // total_parallel_size

        # Calculate ranks within the data parallel group
        dp_rank = group_rank // total_parallel_size

        # Calculate position within the TP-PP grid
        local_rank_in_tp_pp_group = group_rank % total_parallel_size

        # For 2D parallelism: ranks are arranged as [PP0_TP0, PP0_TP1, ..., PP0_TP(tp_size-1), PP1_TP0, ...]
        pp_rank = local_rank_in_tp_pp_group // tp_size
        tp_rank = local_rank_in_tp_pp_group % tp_size

        return dp_size, dp_rank, tp_rank, tp_size, pp_rank, pp_size

    def get_zeromq_address(self):
        return self.zmq_address

    def multi_agent_put_log(self, key: str, data: TensorDict, agent_group: int, next_dp_size: int, timing_raw):
        # This logic needs to be adapted to the new model. For now, it's a warning.
        logger.warning("`multi_agent_put_log` is not yet refactored for DataCoordinator and is a no-op.")
        pass

    def check_mode(self):
        return self.rollout_mode == 'sync' and self._multi_agent == False


# Copyright 2025, Shanghai Innovation Institute. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Checkpoint save/load operations for distributed training."""

import os
import torch
import torch.distributed as dist
from typing import Dict, Optional, Any
from loguru import logger

from siirl.execution.dag.node import NodeRole, NodeType
from siirl.params import SiiRLArguments
from siirl.dag_worker.constants import DAGConstants
from siirl.dag_worker.dag_utils import generate_node_worker_key
from siirl.utils.checkpoint.checkpoint_manager import find_latest_ckpt_path


class CheckpointManager:
    """Manages distributed checkpoint save/load with atomic commits."""

    def __init__(
        self,
        config: SiiRLArguments,
        rank: int,
        gather_group: dist.ProcessGroup,
        workers: Dict[str, Any],
        taskgraph: Any,
        dataloader: Any,
        first_rollout_node: Any,
        get_node_dp_info_fn: callable
    ):
        self.config = config
        self.rank = rank
        self.gather_group = gather_group
        self.workers = workers
        self.taskgraph = taskgraph
        self.dataloader = dataloader
        self.first_rollout_node = first_rollout_node
        self._get_node_dp_info = get_node_dp_info_fn

    def save_checkpoint(self, global_steps: int) -> None:
        """Save checkpoint atomically across all ranks."""
        step_dir = os.path.join(self.config.trainer.default_local_dir, f"global_step_{global_steps}")
        os.makedirs(step_dir, exist_ok=True)
        dist.barrier(self.gather_group)

        logger.info(f"Rank {self.rank}: Saving checkpoint for global_step {global_steps} to {step_dir}")

        self._save_model_states(global_steps, step_dir)
        self._save_dataloader_state(step_dir)

        logger.debug(f"Rank {self.rank}: All data saved. Waiting at barrier before committing checkpoint.")
        dist.barrier(self.gather_group)

        if self.rank == 0:
            self._commit_checkpoint(global_steps)

        dist.barrier(self.gather_group)
        logger.info(f"Rank {self.rank}: Finished saving and committing checkpoint for step {global_steps}.")

    def _save_model_states(self, global_steps: int, step_dir: str) -> None:
        """Save model states for all trainable nodes."""
        saved_worker_keys = set()

        for node in self.taskgraph.nodes.values():
            if node.node_type != NodeType.MODEL_TRAIN:
                continue
            if node.node_role not in [NodeRole.ACTOR, NodeRole.CRITIC]:
                continue

            node_worker_key = generate_node_worker_key(node)

            if node_worker_key in saved_worker_keys:
                continue

            worker = self.workers[node_worker_key]

            sub_dir_name = f"{node.node_role.name.lower()}_agent_{node.agent_group}"
            checkpoint_path = os.path.join(step_dir, sub_dir_name)

            role_name_for_config = node.node_role.name.lower()
            max_ckpt_keep = getattr(
                self.config.trainer,
                f"max_{role_name_for_config}_ckpt_to_keep",
                10
            )

            worker.save_checkpoint(
                local_path=checkpoint_path,
                global_step=global_steps,
                max_ckpt_to_keep=max_ckpt_keep
            )
            saved_worker_keys.add(node_worker_key)
            logger.debug(
                f"Rank {self.rank}: Saved {node.node_role.name} checkpoint for agent {node.agent_group} "
                f"to {checkpoint_path}"
            )

    def _save_dataloader_state(self, step_dir: str) -> None:
        """Save dataloader state (only TP rank 0 and PP rank 0 per DP group)."""
        _, dp_rank, tp_rank, _, pp_rank, _ = self._get_node_dp_info(self.first_rollout_node)

        if tp_rank == 0 and pp_rank == 0:
            dataloader_path = os.path.join(step_dir, f"data_dp_rank_{dp_rank}.pt")
            dataloader_state = self.dataloader.state_dict()
            torch.save(dataloader_state, dataloader_path)
            logger.debug(
                f"Rank {self.rank} (DP_Rank {dp_rank}, TP_Rank {tp_rank}, PP_Rank {pp_rank}): "
                f"Saved dataloader state to {dataloader_path}"
            )

    def _commit_checkpoint(self, global_steps: int) -> None:
        """Atomically commit checkpoint by writing tracker file (rank 0 only)."""
        tracker_file = os.path.join(
            self.config.trainer.default_local_dir,
            "latest_checkpointed_iteration.txt"
        )
        with open(tracker_file, "w") as f:
            f.write(str(global_steps))
        logger.info(f"Rank 0: Checkpoint for step {global_steps} successfully committed.")

    def load_checkpoint(self) -> int:
        """Load checkpoint and return global step to resume from."""
        if self.config.trainer.resume_mode == "disable":
            if self.rank == 0:
                logger.info("Checkpoint loading is disabled. Starting from scratch.")
            return 0

        checkpoint_path = self._determine_checkpoint_path()

        checkpoint_path_container = [checkpoint_path]
        dist.broadcast_object_list(checkpoint_path_container, src=0)
        global_step_folder = checkpoint_path_container[0]

        if global_step_folder is None:
            if self.rank == 0:
                logger.info("No valid checkpoint to load. Training will start from step 0.")
            dist.barrier(self.gather_group)
            return 0

        try:
            global_steps = int(os.path.basename(global_step_folder).split("global_step_")[-1])
            logger.info(
                f"Rank {self.rank}: Resuming from checkpoint. "
                f"Setting global_steps to {global_steps}."
            )
        except (ValueError, IndexError) as e:
            raise ValueError(
                f"Could not parse global step from checkpoint path: {global_step_folder}"
            ) from e

        self._load_model_states(global_step_folder)
        self._load_dataloader_state(global_step_folder)

        dist.barrier(self.gather_group)
        logger.info(f"Rank {self.rank}: Finished loading all checkpoint components.")

        return global_steps

    def _determine_checkpoint_path(self) -> Optional[str]:
        """Determine checkpoint path (rank 0 only)."""
        if self.rank != 0:
            return None

        checkpoint_dir = self.config.trainer.default_local_dir
        resume_from_path = self.config.trainer.resume_from_path
        path_to_load = None

        if self.config.trainer.resume_mode == "auto":
            latest_path = find_latest_ckpt_path(checkpoint_dir)
            if latest_path:
                logger.info(f"Rank 0: Auto-found latest checkpoint at {latest_path}")
                path_to_load = latest_path
        elif self.config.trainer.resume_mode == "resume_path" and resume_from_path:
            logger.info(f"Rank 0: Attempting to load from specified path: {resume_from_path}")
            path_to_load = resume_from_path

        if path_to_load and os.path.exists(path_to_load):
            return path_to_load
        else:
            logger.warning(
                f"Rank 0: Checkpoint path not found or invalid: '{path_to_load}'. "
                f"Starting from scratch."
            )
            return None

    def _load_model_states(self, global_step_folder: str) -> None:
        """Load model states for all trainable nodes."""
        loaded_worker_keys = set()

        for node in self.taskgraph.nodes.values():
            if node.node_type != NodeType.MODEL_TRAIN:
                continue
            if node.node_role not in [NodeRole.ACTOR, NodeRole.CRITIC]:
                continue

            node_worker_key = generate_node_worker_key(node)

            if node_worker_key in loaded_worker_keys:
                continue

            worker = self.workers[node_worker_key]

            sub_dir_name = f"{node.node_role.name.lower()}_agent_{node.agent_group}"
            checkpoint_path = os.path.join(global_step_folder, sub_dir_name)

            if os.path.exists(checkpoint_path):
                worker.load_checkpoint(
                    local_path=checkpoint_path,
                    del_local_after_load=self.config.trainer.del_local_ckpt_after_load
                )
                loaded_worker_keys.add(node_worker_key)
                logger.debug(
                    f"Rank {self.rank}: Loaded {node.node_role.name} checkpoint for agent "
                    f"{node.agent_group} from {checkpoint_path}"
                )
            else:
                logger.warning(
                    f"Rank {self.rank}: Checkpoint for agent {node.agent_group}'s "
                    f"{node.node_role.name} not found at {checkpoint_path}. "
                    f"Weights will be from initialization. "
                    f"If has multi-agent, will share the same checkpoint in agents"
                )

    def _load_dataloader_state(self, global_step_folder: str) -> None:
        """Load dataloader state for current DP group."""
        _, dp_rank, _, _, _, _ = self._get_node_dp_info(self.first_rollout_node)
        dataloader_path = os.path.join(global_step_folder, f"data_dp_rank_{dp_rank}.pt")

        if os.path.exists(dataloader_path):
            dataloader_state = torch.load(dataloader_path, map_location="cpu")
            self.dataloader.load_state_dict(dataloader_state)
            logger.debug(
                f"Rank {self.rank} (DP_Rank {dp_rank}): Loaded dataloader state from "
                f"{dataloader_path}"
            )
        else:
            logger.warning(
                f"Rank {self.rank} (DP_Rank {dp_rank}): Dataloader checkpoint not found at "
                f"{dataloader_path}. Sampler state will not be restored, which may lead to "
                f"data inconsistency."
            )



# Copyright 2025, Shanghai Innovation Institute. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
from contextlib import contextmanager
import os
import time
import multiprocessing
from collections import defaultdict
from datetime import datetime

import numpy as np
from siirl.params import ActorRolloutRefArguments
import tensorflow as tf
import torch
import torch.distributed
from PIL import Image
from loguru import logger
from tensordict import TensorDict
from torch import nn
from torch.nn.utils.rnn import pad_sequence
from torch.distributed.fsdp import FullyShardedDataParallel as FSDP
from transformers import AutoProcessor, GenerationConfig

from siirl.utils.model_utils.torch_functional import get_eos_mask
import siirl.utils.model_utils.torch_functional as siirl_F
from siirl.engine.rollout.base import BaseRollout
from siirl.utils.embodied.video_emb import VideoEmbeddingModel


if multiprocessing.get_start_method(allow_none=True) != "spawn":
    multiprocessing.set_start_method("spawn", force=True)

from siirl.environment.embodied import LIBEROAdapter


__all__ = ['RobHFRollout']

OPENVLA_V01_SYSTEM_PROMPT = (
    "A chat between a curious user and an artificial intelligence assistant. "
    "The assistant gives helpful, detailed, and polite answers to the user's questions."
)


@contextmanager
def _timer(name: str, timing_dict: dict):
    """A context manager to measure execution time of a code block."""
    start_time = time.perf_counter()
    yield
    end_time = time.perf_counter()
    timing_dict[name] = timing_dict.get(name, 0) + end_time - start_time


def crop_and_resize(image, crop_scale, batch_size):
    """
    Center-crops an image to have area `crop_scale` * (original image area), and then resizes back
    to original size. We use the same logic seen in the `dlimp` RLDS datasets wrapper to avoid
    distribution shift at test time.

    Args:
        image: TF Tensor of shape (batch_size, H, W, C) or (H, W, C) and datatype tf.float32 with
               values between [0,1].
        crop_scale: The area of the center crop with respect to the original image.
        batch_size: Batch size.
    """
    # Convert from 3D Tensor (H, W, C) to 4D Tensor (batch_size, H, W, C)
    assert image.shape.ndims == 3 or image.shape.ndims == 4
    expanded_dims = False
    if image.shape.ndims == 3:
        image = tf.expand_dims(image, axis=0)
        expanded_dims = True

    # Get height and width of crop
    new_heights = tf.reshape(tf.clip_by_value(tf.sqrt(crop_scale), 0, 1), shape=(batch_size,))
    new_widths = tf.reshape(tf.clip_by_value(tf.sqrt(crop_scale), 0, 1), shape=(batch_size,))

    # Get bounding box representing crop
    height_offsets = (1 - new_heights) / 2
    width_offsets = (1 - new_widths) / 2
    bounding_boxes = tf.stack(
        [
            height_offsets,
            width_offsets,
            height_offsets + new_heights,
            width_offsets + new_widths,
        ],
        axis=1,
    )

    # Crop and then resize back up
    image = tf.image.crop_and_resize(image, bounding_boxes, tf.range(batch_size), (224, 224))

    # Convert back to 3D Tensor (H, W, C)
    if expanded_dims:
        image = image[0]

    return image


def center_crop_image(image):
    batch_size = 1
    crop_scale = 0.9

    # Convert to TF Tensor and record original data type (should be tf.uint8)
    image = tf.convert_to_tensor(np.array(image))
    orig_dtype = image.dtype

    # Convert to data type tf.float32 and values between [0,1]
    image = tf.image.convert_image_dtype(image, tf.float32)

    # Crop and then resize back to original size
    image = crop_and_resize(image, crop_scale, batch_size)

    # Convert back to original data type
    image = tf.clip_by_value(image, 0, 1)
    image = tf.image.convert_image_dtype(image, orig_dtype, saturate=True)

    # Convert back to PIL Image
    image = Image.fromarray(image.numpy())
    image = image.convert("RGB")
    return image


class EmbodiedHFRollout(BaseRollout):

    def __init__(self, module: nn.Module, config: ActorRolloutRefArguments):
        super().__init__()
        self.config = config
        self.model = module
        self.max_steps = {
            "libero_spatial": 512,
            "libero_object": 512,
            "libero_goal": 512,
            "libero_10": 512,
            "libero_90": 512
        }
        self._embodied_processed = False

        self._rank = torch.distributed.get_rank(
        ) if torch.distributed.is_initialized() else 0
        self._num_gpus_per_node = self.config.embodied.n_gpus_per_node

        self.embedding_model = VideoEmbeddingModel(
            model_path=config.embodied.video_embedding_model_path,
            img_size=config.embodied.embedding_img_size,
            enable_fp16=config.embodied.embedding_enable_fp16
        )

        self.enable_perf = os.environ.get("SIIRL_ENABLE_PERF", "0") == "1"
        self.embedding_model_offload = config.embodied.embedding_model_offload

        # Initialize LIBEROAdapter
        self.num_workers = self.config.embodied.env.num_envs
        # Distribute workers across available GPUs based on rank
        self.adapter = LIBEROAdapter(
            env_name=self.config.embodied.env.env_name,
            num_envs=self.num_workers,
            max_steps=self.config.embodied.env.max_steps,
            num_steps_wait=self.config.embodied.env.num_steps_wait,
            model_family=self.config.embodied.env.model_family,
            gpu_ids=[self._rank % self._num_gpus_per_node] # Run all workers on the same assigned GPU
        )
        logger.info(
            f"Initializing LIBEROAdapter with {self.num_workers} environments...")

    def close(self):
        """Gracefully shuts down the environment adapter."""
        logger.info("Closing LIBEROAdapter...")
        if hasattr(self, 'adapter') and self.adapter:
            self.adapter.close()
        logger.info("LIBEROAdapter closed.")

    def __del__(self):
        # Ensure workers are closed when the object is garbage collected
        self.close()

    def embodied_preprocess(self):
        self.processor = AutoProcessor.from_pretrained(self.config.model.path, trust_remote_code=True)

        if self.config.embodied.embodied_type in ["openvla", "openvla-oft"]:
            gpus = tf.config.experimental.list_physical_devices('GPU')
            if gpus:
                for gpu in gpus:
                    tf.config.experimental.set_memory_growth(gpu, True)

        if self.config.embodied.embodied_type in ["openvla-oft"]:
            if self.config.embodied.unnorm_key not in self.model.norm_stats and f"{self.config.embodied.unnorm_key}_no_noops" in self.model.norm_stats:
                self.config.embodied.unnorm_key = f"{self.config.embodied.unnorm_key}_no_noops"
            assert self.config.embodied.unnorm_key in self.model.norm_stats, f"Action un-norm key {self.config.embodied.unnorm_key} not found in VLA `norm_stats`!"

    def generate_sequences(self, prompts):
        """
        Main entry point for generating sequences.
        It splits a large batch of prompts into chunks that fit the number of workers,
        processes each chunk to generate a rollout, and then concatenates the results.
        This mimics the behavior of the original script to ensure data format compatibility.
        """
        # Preprocess the VLA model only once
        if not self._embodied_processed:
            self.embodied_preprocess()
            self._embodied_processed = True

        tic = time.time()

        total_batch_size = prompts.batch_size[0]
        n_samples = prompts['n_samples'] if 'n_samples' in prompts else 1
        assert self.num_workers >= n_samples, f"rollout num_workers({self.num_workers}) must be >= n_samples({n_samples})"
        batch_size_per_chunk = self.num_workers // n_samples
        num_chunks = (total_batch_size + batch_size_per_chunk - 1) // batch_size_per_chunk
        logger.info(f"RobHFRollout.generate_sequences called with total batch size {total_batch_size}, "
                    f"n_samples {n_samples}, num_workers {self.num_workers}, batch_size_per_chunk {batch_size_per_chunk}, "
                    f"num_chunks {num_chunks}")
        
        all_chunk_outputs = []

        for i in range(num_chunks):
            start_idx = i * batch_size_per_chunk
            end_idx = min((i + 1) * batch_size_per_chunk, total_batch_size)
            
            # Slice the prompts to create a chunk
            chunk_prompts = prompts[start_idx:end_idx]
            
            logger.info(
                f"--- Processing chunk {i+1}/{num_chunks}, size = {chunk_prompts.batch_size[0]} ---")
            
            # Process one chunk and get its TensorDict output
            chunk_output = self._generate_chunk_rollout(chunk_prompts)
            all_chunk_outputs.append(chunk_output)

        # Concatenate the TensorDict objects from all chunks
        final_output = torch.cat(all_chunk_outputs)
        logger.info(f"RobHFRollout.generate_sequences finished for a single batch of size {final_output.batch_size[0]}"
                    f", took {time.time() - tic:.2f} seconds")
        return final_output

    def process_input(self,inputs:list, task_descriptions:list):
        
        batchdata = {"input_ids":[],"attention_mask":[],"pixel_values":[]}  
        
        for i in range(len(inputs)):
            input = inputs[i]
            task_description = task_descriptions[i]
           
            image = Image.fromarray(input["full_image"]).convert("RGB")
            if self.config.embodied.center_crop:
                image = center_crop_image(image)
            prompt = f"In: What action should the robot take to {task_description.lower()}?\nOut:"
            batch_feature  = self.processor(prompt, image)
            
            if "wrist_image" in input.keys():
                wrist_image = Image.fromarray(input["wrist_image"]).convert("RGB")
                if self.config.embodied.center_crop:
                    wrist_image = center_crop_image(wrist_image)
                wrist_batch_feature = self.processor(prompt, wrist_image)
                primary_pixel_values = batch_feature["pixel_values"]
                batch_feature["pixel_values"] = torch.cat([primary_pixel_values] + [wrist_batch_feature["pixel_values"]], dim=1)
                
            input_ids = batch_feature["input_ids"]
            attention_mask = batch_feature["attention_mask"]
            pixel_values = batch_feature["pixel_values"]
            
            if not torch.all(input_ids[:, -1] == 29871):
                input_ids = torch.cat(
                    (input_ids, torch.unsqueeze(torch.Tensor([29871]).long(), dim=0).to(input_ids.device)), dim=1
                )
                if self.config.embodied.embodied_type in ["openvla-oft"]:
                    attention_mask = torch.cat(
                        (attention_mask, torch.unsqueeze(torch.Tensor([True]).bool(), dim=0).to(attention_mask.device)), dim=1
                    )
            
            batchdata["input_ids"].append(input_ids)    
            batchdata["attention_mask"].append(attention_mask)    
            batchdata["pixel_values"].append(pixel_values)    
        
        
        device = torch.device('cuda') 
        
        if self.config.embodied.embodied_type in ["openvla-oft"]:
            batchdata["input_ids"] = [x.transpose(0, 1) for x in batchdata["input_ids"]]
            batchdata["attention_mask"] = [x.transpose(0, 1) for x in batchdata["attention_mask"]]
            batchdata["input_ids"] = pad_sequence(batchdata["input_ids"], batch_first=True, padding_value=self.processor.tokenizer.pad_token_id).squeeze(-1).to(device)
            batchdata["attention_mask"] = pad_sequence(batchdata["attention_mask"], batch_first=True, padding_value=0).squeeze(-1).to(device)
            
            padding_mask = batchdata["input_ids"].ne(self.processor.tokenizer.pad_token_id)
            assert  torch.all(padding_mask==batchdata["attention_mask"].ne(0))
            padding_mask = ~padding_mask
            padding_mask = padding_mask.int() 
            sorted_indices = torch.argsort(padding_mask, dim=1, descending=True, stable=True)
            batchdata["input_ids"] = torch.gather(batchdata["input_ids"], 1, sorted_indices)
            batchdata["attention_mask"] = torch.gather(batchdata["attention_mask"], 1, sorted_indices)
            
            
            batchdata["pixel_values"] = torch.cat(batchdata["pixel_values"] , dim=0).to(device)
            assert torch.all(batchdata["attention_mask"].ne(0) == batchdata["input_ids"].ne(self.processor.tokenizer.pad_token_id))
        else:
            for key in ["input_ids", "attention_mask", "pixel_values"]:
                batchdata[key] = torch.cat(batchdata[key], dim=0).to(device)

        return batchdata

    def _generate_chunk_rollout(self, prompts):
        generate_tic = time.time()
        self.model.eval()
        # n_samples = prompts.get('n_samples', 1)
        task_id = prompts['task_id']
        trial_id = prompts['trial_id']
        task_suite_name = prompts['task_suite_name']
        assert np.all(task_suite_name == self.config.embodied.env.env_name), \
            "All task_suite_name in the batch must match the rollout config"
        max_steps = self.config.embodied.env.max_steps
        chunk_size = task_id.size(0)

        is_valid = "n_samples" in prompts
        global_steps = prompts.get('global_steps', 0) if is_valid else 0

        timing_dict = {}

        # Reset environments using the adapter
        with _timer(f"adapter_reset", timing_dict):
            # This is a blocking call
            init_data_list = self.adapter._blocking_reset(
                task_ids=task_id.reshape(-1).cpu().numpy().tolist(),
                trial_ids=trial_id.reshape(-1).cpu().numpy().tolist(),
            )

        inputs = [None] * chunk_size
        task_descriptions = [None] * chunk_size
        task_records = [None] * chunk_size
        valid_video = defaultdict(list)
        all_video = defaultdict(list)

        # Collect initial observations for the chunk
        with _timer(f"process_initial_obs", timing_dict):
            for idx in range(chunk_size):
                init_data = init_data_list[idx]
                task_descriptions[idx] = init_data["task_description"]
                inputs[idx] = self._obs_to_input(init_data['obs'])
                task_records[idx] = {
                    "active": init_data['active'],
                    "complete": init_data['complete'],
                    "finish_step": init_data['finish_step'],
                    "task_file_name": init_data['task_file_name']
                }
                if is_valid:
                    valid_video[init_data['task_file_name']].extend(
                        init_data['valid_images'])
                all_video[init_data['task_file_name']].extend(
                    init_data['valid_images'])

        step = 0
        vla_history = []
        meta_info_keys = ["eos_token_id", "pad_token_id", "recompute_log_prob", "validate", "do_sample", "global_steps"]
        meta_info_keys = [key for key in meta_info_keys if key in prompts.keys()]
        meta_info = prompts.select(*meta_info_keys)
        while step < max_steps:
            active_indices = [i for i, r in enumerate(task_records) if r['active']]
            current_inputs = inputs
            current_task_descriptions = task_descriptions

            with _timer(f"process_input", timing_dict):
                vla_input = self.process_input(current_inputs, current_task_descriptions)
            vla_input.update(meta_info)

            with _timer(f"_generate_one_step", timing_dict):
                vla_output = self._generate_one_step(vla_input)
            
            actions = vla_output["action"]

            step_data = {
                "responses": vla_output["responses"],
                "input_ids": vla_output["input_ids"],
                "attention_mask": vla_output["attention_mask"],
                "pixel_values": vla_output["pixel_values"],
                "action": actions,
                "step": step
            }
            vla_history.append(step_data)

            with _timer(f"adapter_step", timing_dict):
                step_results_list = self.adapter._blocking_step({
                    "indices": active_indices,
                    "actions": actions,
                })
            
            with _timer(f"process_step_results", timing_dict):
                new_inputs = inputs.copy()
                for idx in active_indices:
                    result = step_results_list[idx]
                    new_inputs[idx] = self._obs_to_input(result['obs'])
                    task_records[idx]['active'] = result['active']
                    task_records[idx]['complete'] = result['complete']
                    task_records[idx]['finish_step'] = result['finish_step']
                    all_video[task_records[idx]['task_file_name']].extend(result['valid_images'])
                    if is_valid:
                        valid_video[task_records[idx]['task_file_name']].extend(result['valid_images'])
                inputs = new_inputs
            
            step += self.config.embodied.action_chunks_len
        
        with _timer(f"post_loop_processing", timing_dict):
            torch.cuda.empty_cache()
            self.model.train()
            
            batch = {
                    'responses': [],
                    'input_ids': [],  # here input_ids become the whole sentences
                    'attention_mask': [],
                    'pixel_values': [],
                }
            for k in ["responses", "input_ids", "attention_mask", "pixel_values"]:
                for h in vla_history:
                    batch[k].append(h[k])
            
            for k,v in batch.items():
                batch[k] = torch.stack(v,dim=1) 
    
            batch["complete"] = []
            batch["finish_step"] = []
            batch["task_file_name"] = []
    
            for k in task_records:
                batch["complete"].append(k["complete"])
                batch["finish_step"].append(k["finish_step"])
                batch["task_file_name"].append(k["task_file_name"])
            
            batch["complete"] = torch.tensor(batch["complete"], dtype=torch.bool, device=batch['responses'].device)
            batch["finish_step"] = torch.tensor(batch["finish_step"], dtype=torch.int64, device=batch['responses'].device)
            # 构建 batch
            names = batch["task_file_name"]
            max_len = 50 # max(len(n) for n in names)
            padded = [n.ljust(max_len, '\0') for n in names]
            batch["task_file_name"] = torch.tensor(
                [s.encode('utf-8') for s in padded],
                dtype=torch.uint8,
                device=batch['responses'].device
            )

        vjepa_embeddings = []
        tasks_for_embedding = []
        for k in task_records:
            tasks_for_embedding.append((
                k['task_file_name'],
                all_video.get(k['task_file_name'], []),
                "rollout4embedding",
                global_steps,
                k['complete']
            ))
        
        with _timer(f"get_embeddings", timing_dict):
            batch_names, batch_frames = zip(*[(t[0], t[1])  for t in tasks_for_embedding])
            vjepa_embeddings = self.embedding_model.get_embeddings(batch_names, batch_frames)
            batch["vjepa_embedding"] = torch.tensor(
                np.array(vjepa_embeddings), dtype=torch.float32)

        if self.enable_perf:
            generate_chunk_rollout_time = time.time() - generate_tic
            log_str = f"\n--- ⏱️  Chunk Performance (size={chunk_size}) ---\n"
            
            # Sort the dictionary by value in descending order for better readability
            sorted_timing = sorted(timing_dict.items(), key=lambda item: item[1], reverse=True)

            for key, value in sorted_timing:
                log_str += f"  {key}: {value:.4f} seconds\n"
            log_str += f"  _generate_chunk_rollout: {generate_chunk_rollout_time:.4f} seconds\n"
            log_str += f"  total steps in chunk: {step}\n"
            log_str += "--- ⏱️  End of Chunk Performance Log ---\n"

            with open(f"rollout_performance_rank_{self._rank}.log", "a") as f:
                f.write(f"\n{datetime.now()}:\n")
                f.write(log_str)
            logger.info(log_str)
        
        output_batch = TensorDict(
            batch,
            batch_size=chunk_size)
        return output_batch

    @torch.no_grad()
    def _generate_one_step(self, prompts: dict):
        if self.config.embodied.embodied_type == "openvla-oft":
            idx = prompts['input_ids']  # (bs, prompt_length)
            attention_mask = prompts['attention_mask']  # left-padded attention_mask
            pixel_values = prompts["pixel_values"]
        
        
            param_ctx = contextlib.nullcontext()

            # make sampling args can be overriden by inputs
            do_sample = prompts.get('do_sample', self.config.rollout.do_sample)
        

            temperature = prompts.get('temperature', self.config.rollout.temperature)

            #generation_config = GenerationConfig(temperature=temperature, top_p=top_p, top_k=top_k)

            if isinstance(self.model, FSDP):
                # recurse need to set to False according to https://github.com/pytorch/pytorch/issues/100069
                param_ctx = FSDP.summon_full_params(self.model, writeback=False, recurse=False)
            
            with param_ctx:
                with torch.autocast(device_type='cuda', dtype=torch.bfloat16):
                    actions, response = self.model.generate_action_verl(
                        input_ids=idx,
                        pixel_values=pixel_values,
                        attention_mask=attention_mask,
                        padding_idx = self.processor.tokenizer.pad_token_id,
                        do_sample=do_sample,
                        unnorm_key=self.config.embodied.unnorm_key,
                        temperature=temperature, )
            
            
            assert self.processor.tokenizer.pad_token_id is not None

            assert idx.ndim == 2
            idx = siirl_F.pad_sequence_to_length(idx,max_seq_len=self.config.rollout.prompt_length,pad_token_id=self.processor.tokenizer.pad_token_id,left_pad=True)
            
            assert attention_mask.ndim == 2
            attention_mask = siirl_F.pad_sequence_to_length(attention_mask,max_seq_len=self.config.rollout.prompt_length,pad_token_id=0,left_pad=True)
            
            
            assert idx.device.type == 'cuda'
            assert response.device.type == 'cuda'
            #assert seq.device.type == 'cuda'
            assert attention_mask.device.type == 'cuda'
            assert pixel_values.device.type == 'cuda'
            batch ={
                    'responses': response,
                    'input_ids': idx,
                    'attention_mask': attention_mask,
                    "pixel_values":pixel_values,
                    "action":actions,
                }

            return batch
        
        elif self.config.embodied.embodied_type == "openvla": 
            idx = prompts['input_ids']  # (bs, prompt_length)
            attention_mask = prompts['attention_mask']  # left-padded attention_mask
            pixel_values = prompts["pixel_values"]
            
            # used to construct attention_mask
            eos_token_id = prompts['eos_token_id']
            pad_token_id = prompts['pad_token_id']

            batch_size = idx.size(0)
            prompt_length = idx.size(1)
            #self.model.eval()
            param_ctx = contextlib.nullcontext()

            do_sample = prompts.get('do_sample', self.config.rollout.do_sample)
            response_length =  self.model.get_action_dim(self.config.embodied.unnorm_key)
            top_p = prompts.get('top_p', self.config.rollout.top_p)
            top_k = prompts.get('top_k', self.config.rollout.top_k)
            if top_k is None:
                top_k = 0
            top_k = max(0, top_k)  # to be compatible with vllm

            temperature = prompts.get('temperature', self.config.rollout.temperature)
            generation_config = GenerationConfig(temperature=temperature, top_p=top_p, top_k=top_k)

            if isinstance(self.model, FSDP):
                # recurse need to set to False according to https://github.com/pytorch/pytorch/issues/100069
                param_ctx = FSDP.summon_full_params(self.model, writeback=False, recurse=False)
            
            with param_ctx:
                with torch.autocast(device_type='cuda', dtype=torch.bfloat16):
                    
                    output = self.model.generate(
                        input_ids=idx,
                        pixel_values=pixel_values,
                        attention_mask=attention_mask,
                        do_sample=do_sample,
                        max_new_tokens=response_length,
                        # max_length=max_length,
                        eos_token_id=eos_token_id,
                        pad_token_id=pad_token_id,
                        generation_config=generation_config,
                        # renormalize_logits=True,
                        output_scores=False,  # this is potentially very large
                        return_dict_in_generate=True,
                        use_cache=True)
                    
           
            seq = output.sequences
            sequence_length = prompt_length + response_length
            delta_length = sequence_length - seq.shape[1]
            
            assert delta_length == 0
            assert seq.shape[1] == sequence_length

            prompt = seq[:, :prompt_length]  # (bs, prompt_length)
            response = seq[:, prompt_length:]  # (bs, response_length)

            response_length = response.size(1)
            #delta_position_id = torch.arange(1, response_length + 1, device=position_ids.device)
            #delta_position_id = delta_position_id.unsqueeze(0).repeat(batch_size, 1)
            #response_position_ids = position_ids[:, -1:] + delta_position_id
            #position_ids = torch.cat([position_ids, response_position_ids], dim=-1)

            response_attention_mask = get_eos_mask(response_id=response, eos_token=eos_token_id, dtype=attention_mask.dtype)
            attention_mask = torch.cat((attention_mask, response_attention_mask), dim=-1)

            # Extract predicted action tokens and translate into (normalized) continuous actions
            predicted_action_token_ids = response.detach().cpu().numpy()
            discretized_actions = self.model.vocab_size - predicted_action_token_ids
            discretized_actions = np.clip(discretized_actions - 1, a_min=0, a_max=self.model.bin_centers.shape[0] - 1)
            normalized_actions = self.model.bin_centers[discretized_actions]

            # Unnormalize actions
            action_norm_stats = self.model.get_action_stats(self.config.embodied.unnorm_key)
            mask = action_norm_stats.get("mask", np.ones_like(action_norm_stats["q01"], dtype=bool))
            action_high, action_low = np.array(action_norm_stats["q99"]), np.array(action_norm_stats["q01"])
            actions = np.where(
                mask,
                0.5 * (normalized_actions + 1) * (action_high - action_low) + action_low,
                normalized_actions,
            )
            
            actions = np.expand_dims(actions, axis=1)
            
            assert self.processor.tokenizer.pad_token_id is not None
            assert prompt.ndim == 2
            prompt = siirl_F.pad_sequence_to_length(prompt,max_seq_len=self.config.rollout.prompt_length,pad_token_id=self.processor.tokenizer.pad_token_id,left_pad=True)
            assert seq.ndim == 2
            seq = siirl_F.pad_sequence_to_length(seq,max_seq_len=self.config.rollout.prompt_length,pad_token_id=self.processor.tokenizer.pad_token_id,left_pad=True)
            assert attention_mask.ndim == 2
            attention_mask = siirl_F.pad_sequence_to_length(attention_mask,max_seq_len=self.config.rollout.prompt_length,pad_token_id=0,left_pad=True)
            
            batch ={
                    'prompts': prompt,
                    'responses': response,
                    'input_ids': seq,
                    'attention_mask': attention_mask,
                    "pixel_values":pixel_values,
                    "action":actions,
                    #'position_ids': position_ids
                }
            
            return batch

    def _obs_to_input(self, obs):
        from siirl.utils.embodied.libero_utils import get_libero_image, get_libero_wrist_image, quat2axisangle

        if self.config.embodied.num_images_in_input > 1:
            return {
                "full_image": get_libero_image(obs, 224),
                "wrist_image": get_libero_wrist_image(obs, 224),
                "state": np.concatenate([
                    obs["robot0_eef_pos"],
                    quat2axisangle(obs["robot0_eef_quat"]),
                    obs["robot0_gripper_qpos"]
                ])
            }
        else:
            return {
                "full_image": get_libero_image(obs, 224),
                "state": np.concatenate([
                    obs["robot0_eef_pos"],
                    quat2axisangle(obs["robot0_eef_quat"]),
                    obs["robot0_gripper_qpos"]
                ])
            }



"""
Unit tests for DAPO filtering and merging logic.
This test simulates the real scenario from training logs to quickly verify fixes.
"""
import torch
import numpy as np
from tensordict import TensorDict
from tensordict.tensorclass import NonTensorData
from siirl.data_coordinator.sample import filter_tensordict


def get_unpacked_data(value):
    """Safely get data from a value that might be NonTensorData or a raw type."""
    if isinstance(value, NonTensorData):
        return value.data
    return value


def create_mock_batch(batch_size=1024):
    """
    Create a mock batch that simulates the real data structure in DAPO.
    
    Includes:
    - Batched tensor fields (input_ids, attention_mask, etc.)
    - Batched NonTensorData fields (data_source, ability, etc.)
    - Metadata NonTensorData fields (eos_token_id, pad_token_id, etc.)
    """
    batch_dict = {
        # Tensor fields (batched)
        'input_ids': torch.randint(0, 1000, (batch_size, 10240)),
        'attention_mask': torch.ones(batch_size, 10240, dtype=torch.int64),
        'position_ids': torch.arange(10240).unsqueeze(0).repeat(batch_size, 1),
        'uid': torch.arange(batch_size, dtype=torch.int64),
        'prompts': torch.randint(0, 1000, (batch_size, 2048)),
        'responses': torch.randint(0, 1000, (batch_size, 8192)),
        'response_mask': torch.ones(batch_size, 8192, dtype=torch.int64),
        'token_level_scores': torch.randn(batch_size, 8192),
        'score': torch.randn(batch_size),
        'acc': torch.randint(0, 2, (batch_size,), dtype=torch.bool),
        'token_level_rewards': torch.randn(batch_size, 8192),
        
        # Batched NonTensorData fields (length = batch_size)
        'data_source': NonTensorData(
            data=np.array(['gsm8k'] * batch_size),
            batch_size=[batch_size]
        ),
        'ability': NonTensorData(
            data=np.array(['math'] * batch_size),
            batch_size=[batch_size]
        ),
        'reward_model': NonTensorData(
            data=np.array([{'name': 'rm1'}] * batch_size),
            batch_size=[batch_size]
        ),
        'extra_info': NonTensorData(
            data=np.array([{}] * batch_size),
            batch_size=[batch_size]
        ),
        'pred': NonTensorData(
            data=np.array(['answer'] * batch_size),
            batch_size=[batch_size]
        ),
        'global_token_num': NonTensorData(
            data=[10240] * batch_size,
            batch_size=[batch_size]
        ),
        
        # Metadata NonTensorData fields (length != batch_size)
        'eos_token_id': NonTensorData(
            data=[151643, 151645],  # Length 2, not batch_size
            batch_size=[batch_size]
        ),
        'pad_token_id': NonTensorData(
            data=151643,  # Scalar
            batch_size=[batch_size]
        ),
        'total_input_tokens': NonTensorData(
            data=168784,  # Scalar
            batch_size=[batch_size]
        ),
        'total_output_tokens': NonTensorData(
            data=966975,  # Scalar
            batch_size=[batch_size]
        ),
    }
    
    return TensorDict(batch_dict, batch_size=batch_size)


def test_filter_tensordict():
    """Test that filter_tensordict correctly handles batched data and metadata."""
    print("\n=== Testing filter_tensordict ===")
    
    # Create mock batch
    batch = create_mock_batch(batch_size=1024)
    print(f"Original batch size: {len(batch)}")
    
    # Simulate filtering (keeping ~50% of samples, like in the logs)
    indices = list(range(16, 1024, 2))  # Keep every other sample starting from 16
    print(f"Filtering to {len(indices)} samples")
    
    # Filter
    filtered_batch = filter_tensordict(batch, indices)
    print(f"Filtered batch size: {len(filtered_batch)}")
    
    # Verify tensor fields are correctly filtered
    assert len(filtered_batch['input_ids']) == len(indices)
    assert len(filtered_batch['uid']) == len(indices)
    
    # Verify batched NonTensorData fields are correctly filtered
    assert len(get_unpacked_data(filtered_batch['data_source'])) == len(indices)
    assert len(get_unpacked_data(filtered_batch['ability'])) == len(indices)
    assert len(get_unpacked_data(filtered_batch['global_token_num'])) == len(indices)
    
    # Verify metadata NonTensorData fields are preserved (not filtered)
    assert len(get_unpacked_data(filtered_batch['eos_token_id'])) == 2, "Metadata should be preserved"
    assert get_unpacked_data(filtered_batch['pad_token_id']) == 151643, "Metadata should be preserved"
    
    print("✅ filter_tensordict test passed!")
    return filtered_batch


def test_merge_cached_and_new():
    """
    Test merging cached filtered batch with new filtered batch.
    This simulates the real merge scenario in dagworker.py.
    """
    print("\n=== Testing merge logic ===")
    
    # Create first batch and filter (this will be cached)
    batch1 = create_mock_batch(batch_size=1024)
    indices1 = list(range(16, 512))  # 496 samples
    cached_batch = filter_tensordict(batch1, indices1)
    print(f"Cached batch size: {len(cached_batch)}")
    
    # Create second batch and filter (this is the new batch)
    batch2 = create_mock_batch(batch_size=1024)
    indices2 = list(range(20, 540))  # 520 samples
    new_batch = filter_tensordict(batch2, indices2)
    print(f"New batch size: {len(new_batch)}")
    
    # Now merge them (simulate dagworker merge logic)
    cache_size = len(cached_batch)
    new_size = len(new_batch)
    merged_size = cache_size + new_size
    print(f"Expected merged size: {merged_size}")
    
    merged_dict = {}
    for key in cached_batch.keys():
        cache_val = cached_batch[key]
        new_val = new_batch[key]
        
        # Debug: print info for NonTensorData fields
        if isinstance(cache_val, NonTensorData) and key in ['data_source', 'ability', 'eos_token_id']:
            print(f"\nDEBUG {key}:")
            print(f"  cache_val type: {type(cache_val)}, data type: {type(cache_val.data)}")
            if hasattr(cache_val.data, '__len__'):
                print(f"  cache_val.data length: {len(cache_val.data)}, cache_size: {cache_size}")
            print(f"  new_val type: {type(new_val)}, data type: {type(new_val.data) if isinstance(new_val, NonTensorData) else 'N/A'}")
            if isinstance(new_val, NonTensorData) and hasattr(new_val.data, '__len__'):
                print(f"  new_val.data length: {len(new_val.data)}, new_size: {new_size}")
        
        if isinstance(cache_val, torch.Tensor):
            # Merge tensors
            merged_dict[key] = torch.cat([cache_val, new_val], dim=0)
        else:
            # Handle NonTensorData or raw types (TensorDict may unpack NonTensorData)
            # Extract the actual data
            if isinstance(cache_val, NonTensorData):
                cache_data = cache_val.data
            else:
                cache_data = cache_val
            
            if isinstance(new_val, NonTensorData):
                new_data = new_val.data
            else:
                new_data = new_val
            
            # Determine if it's batched data or metadata based on length
            if isinstance(cache_data, np.ndarray):
                cache_len = len(cache_data)
                if cache_len == cache_size:
                    # Batched np.ndarray - merge
                    new_arr = new_data if isinstance(new_data, np.ndarray) else np.array(new_data)
                    merged_data = np.concatenate([cache_data, new_arr], axis=0)
                    merged_dict[key] = NonTensorData(data=merged_data, batch_size=[merged_size])
                else:
                    # Metadata - keep as NonTensorData
                    merged_dict[key] = NonTensorData(data=new_data, batch_size=[merged_size])
            elif isinstance(cache_data, (list, tuple)):
                cache_len = len(cache_data)
                if cache_len == cache_size:
                    # Batched list - merge
                    new_list = new_data if isinstance(new_data, (list, tuple)) else [new_data] * new_size
                    merged_data = list(cache_data) + list(new_list)
                    merged_dict[key] = NonTensorData(data=merged_data, batch_size=[merged_size])
                else:
                    # Metadata - keep as NonTensorData
                    merged_dict[key] = NonTensorData(data=new_data, batch_size=[merged_size])
            else:
                # Scalar - always metadata
                merged_dict[key] = NonTensorData(data=new_data, batch_size=[merged_size])
    
    # Verify all fields before creating TensorDict
    print("\n--- Pre-creation validation ---")
    for key, value in merged_dict.items():
        if isinstance(value, torch.Tensor):
            val_len = value.shape[0]
            status = "✅" if val_len == merged_size else "❌"
            print(f"{status} {key}: Tensor, shape={value.shape}")
            assert val_len == merged_size, f"Tensor {key} has wrong size: {val_len} != {merged_size}"
        elif isinstance(value, NonTensorData):
            if hasattr(value.data, '__len__') and not isinstance(value.data, str):
                val_len = len(value.data)
                # For batched data, length should match merged_size
                # For metadata, length can be different
                if val_len == merged_size:
                    print(f"✅ {key}: NonTensorData(batched), len={val_len}")
                else:
                    print(f"✅ {key}: NonTensorData(metadata), len={val_len}")
            else:
                print(f"✅ {key}: NonTensorData(scalar)")
        else:
            print(f"❌ {key}: Raw type {type(value)} - should be wrapped!")
            raise ValueError(f"Field {key} is not properly wrapped")
    
    # Try to create TensorDict
    print("\n--- Creating TensorDict ---")
    try:
        merged_batch = TensorDict(merged_dict, batch_size=merged_size)
        print(f"✅ Successfully created merged TensorDict with batch_size={merged_size}")
        
        # Verify the result
        assert len(merged_batch) == merged_size
        assert len(merged_batch['input_ids']) == merged_size
        assert len(get_unpacked_data(merged_batch['data_source'])) == merged_size
        
        # Verify metadata is preserved
        assert len(get_unpacked_data(merged_batch['eos_token_id'])) == 2
        assert get_unpacked_data(merged_batch['pad_token_id']) == 151643
        
        print("✅ Merge test passed!")
        return merged_batch
        
    except Exception as e:
        print(f"❌ Failed to create TensorDict: {e}")
        
        # Print detailed error info
        print("\n--- Error Details ---")
        for k, v in merged_dict.items():
            v_type = type(v).__name__
            v_shape = getattr(v, 'shape', 'None')
            v_dtype = getattr(v, 'dtype', 'None')
            if isinstance(v, NonTensorData):
                v_shape = f"({len(v.data)},)" if hasattr(v.data, '__len__') else 'scalar'
                v_dtype = type(v.data).__name__
            print(f"  '{k}' -> type={v_type}, shape={v_shape}, dtype={v_dtype}")
        
        raise


def test_full_dapo_workflow():
    """
    Test the complete DAPO workflow:
    1. First rollout + filter -> insufficient samples -> cache
    2. Second rollout + filter -> merge with cache -> sufficient samples
    """
    print("\n=== Testing full DAPO workflow ===")
    
    # Step 1: First rollout
    print("\n--- Step 1: First rollout ---")
    batch1 = create_mock_batch(batch_size=1024)
    indices1 = list(range(16, 512))  # 496 samples (insufficient)
    filtered1 = filter_tensordict(batch1, indices1)
    print(f"First rollout: {len(filtered1)} samples (insufficient)")
    
    # Cache it
    cached_batch = filtered1
    
    # Step 2: Second rollout
    print("\n--- Step 2: Second rollout ---")
    batch2 = create_mock_batch(batch_size=1024)
    indices2 = list(range(20, 540))  # 520 samples
    filtered2 = filter_tensordict(batch2, indices2)
    print(f"Second rollout: {len(filtered2)} samples")
    
    # Step 3: Merge
    print("\n--- Step 3: Merge cached and new ---")
    merged = test_merge_cached_and_new()
    
    print(f"\n✅ Full workflow test passed! Final batch size: {len(merged)}")


if __name__ == '__main__':
    print("=" * 60)
    print("DAPO Merge Logic Unit Tests")
    print("=" * 60)
    
    try:
        # Test 1: Filter
        filtered = test_filter_tensordict()
        
        # Test 2: Merge
        merged = test_merge_cached_and_new()
        
        # Test 3: Full workflow
        test_full_dapo_workflow()
        
        print("\n" + "=" * 60)
        print("✅ ALL TESTS PASSED!")
        print("=" * 60)
        
    except Exception as e:
        print("\n" + "=" * 60)
        print(f"❌ TEST FAILED: {e}")
        print("=" * 60)
        import traceback
        traceback.print_exc()
        exit(1)

